from typing import List, ClassVar
from dataclasses import dataclass, field
from enum import Enum

from fs.base import FS
from fs.path import combine as fs_combine
from fs.path import dirname as fs_dirname
from fs.path import basename as fs_basename

from zetsubou.project.base_context import BaseContext
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.project.runtime.resolve import ResolvedTarget
from zetsubou.project.config_matrix import ConfigMatrix
from zetsubou.commands.command_context import Conan


class EDefaultTargets(Enum):
    All = 0x1
    Update = 0x2


@dataclass
class EmitContext(BaseContext):
    project_template: ProjectTemplate = None
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)
    config_matrix: ConfigMatrix = None
    conan: Conan = None
    mem_fs: FS = None
    ROOT_DIR: ClassVar[str] = "."

    @classmethod
    def from_base(cls, source : BaseContext):
        result = EmitContext()
        result.__dict__.update(source.__dict__)
        return result

    def write_file(self, rw_path: str, content: str):
        dirname = fs_dirname(rw_path)
        filename = fs_basename(rw_path)

        dirname = fs_combine(EmitContext.ROOT_DIR, dirname)

        with self.mem_fs.makedirs(dirname, recreate=True) as out_dir:
            with out_dir.open(filename, 'w') as out_file:
                out_file.write(content)


class Emitter:
    def emit_solution(self, context: EmitContext) -> str:
        raise NotImplementedError()

    def emit_target(self, context: EmitContext, target: ResolvedTarget):
        raise NotImplementedError()


def emit_project(context: EmitContext, emitter: Emitter):
    # emit target files
    for target in context.resolved_targets:
        emitter.emit_target(context, target)

    # emit main project file
    main_file = emitter.emit_solution(context)

    return (context.mem_fs, EmitContext.ROOT_DIR, main_file)



from subprocess import Popen as process_open
from subprocess import PIPE, STDOUT
from sys import stdout as sys_stdout
from typing import List, Callable, Optional

from zetsubou.project.model.virtual_environment import VirtualEnvironment
from zetsubou.utils.common import fix_path_os
from zetsubou.utils.error_codes import EErrorCode


class EnterVenv():
    venv_obj : Optional[VirtualEnvironment] = None

    def __init__(self, venv_obj : VirtualEnvironment):
        self.venv_obj = venv_obj

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def call_process(self, args : List[str], realtime : bool = False, capture : bool = True, on_error : Callable[[int], EErrorCode] = None, cwd: str = '.'):
        return call_process_venv(args, self.venv_obj, realtime, capture, on_error, cwd)


def call_process_venv(args : List[str], venv : VirtualEnvironment, realtime : bool = False, capture : bool = True, on_error : Callable[[int], EErrorCode] = None, cwd: str = '.'):
    return call_process([ fix_path_os(venv.activate), '&&' ] + args, realtime, capture, on_error, cwd)


def call_process(args : List[str], realtime : bool = False, capture : bool = True, on_error : Callable[[int], EErrorCode] = None, cwd: str = '.'):
    str_buff : List[str] = []
    err = None

    with process_open(args, encoding=sys_stdout.encoding, shell=True, stdout=PIPE, stderr=STDOUT, bufsize=1,
                universal_newlines=True, cwd=cwd, errors='ignore') as proc:

        while proc.stdout.readable():
            line = proc.stdout.readline()
            if capture:
                str_buff.append(line)
            if realtime:
                sys_stdout.write(line)

            if line == '' and proc.poll() is not None:
                break

        if proc.returncode != 0:
            if on_error is not None:
                err = on_error(proc.returncode)
            else:
                err = proc.returncode

        return (''.join(str_buff), err)



import os
from dataclasses import dataclass, field
from typing import List, Optional, Any
from fs.base import FS

from zetsubou.commands.execute_stage import execute_stage
from zetsubou.utils import logger
from zetsubou.utils.common import fix_path
from zetsubou.project.config_matrix import ConfigMatrix
from zetsubou.project.base_context import BaseContext
from zetsubou.project.model.target import Target
from zetsubou.project.model.virtual_environment import VirtualEnvironment, resolve_venv
from zetsubou.project.runtime.resolve import ResolvedTarget
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.utils.error_codes import EErrorCode

@dataclass
class Fastbuild:
    emit_fs : FS
    bff_dir : str
    bff_file : str


@dataclass
class Conan:
    yml_files: List[str] = field(default_factory=list)
    dependencies: List[Target] = field(default_factory=list)
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)


@dataclass
class CommandContext(BaseContext):
    command_args: Any = None

    project_template: Optional[ProjectTemplate] = None
    # toolchains: List[ToolchainDefinition] = field(default_factory=list)
    config_matrix: Optional[ConfigMatrix] = None
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)
    fastbuild: Optional[Fastbuild] = None
    conan: Conan = field(default_factory=Conan)

    def to_out_path(self, path : str):
        return fix_path(os.path.join(self.fs_root, path))

    def resolve_cli_tools(self):
        for cli_tool in self.project_template.cli_tools:
            if not cli_tool.resolve(self.fs_venv):
                logger.Error(f"Unable to resolve fullpath to '{cli_tool.executable}'. Might be missing from venv or PATH.")
                return False
        return True

    def resolve_venv(self):
        if self.fs_venv is None:
            venv_obj : VirtualEnvironment = execute_stage(lambda: resolve_venv(self.project_fs, self.project_template.project, self.fs_root),
                                        'Virtual environemnt found',
                                        EErrorCode.UNABLE_TO_FIND_VENV)

            logger.Info(f"Virtual environement - '{venv_obj.activate}'")
            self.fs_venv = VirtualEnvironment(self.to_out_path(venv_obj.activate), self.to_out_path(venv_obj.deactivate))

            # resolve fullpaths to cli tools
            execute_stage(self.resolve_cli_tools,
                          'Commandline tools found',
                          EErrorCode.UNABLE_TO_RESOLVE_CLI_TOOLS_FULLPATH)



from typing import List, Optional
from dataclasses import dataclass, field
from dataclasses import dataclass

from zetsubou.project.model.platform_enums import ESystem, EArch
from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.target import TargetData
from bentoudev.dataclass.base import loaded_from_file


@dataclass
class PlatformData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
@loaded_from_file
class Platform(TargetData):
    platform: str = None
    host_system: ESystem = None
    host_arch: EArch = None
    target_arch: EArch = None
    filters: Optional[List[PlatformData]] = field(default_factory=list)



from typing import List, Optional
import os

from bentoudev.dataclass.yaml_loader import LineLoader

from zetsubou.project.model.virtual_environment import VirtualEnvironment
from zetsubou.utils.subprocess import call_process, call_process_venv
import zetsubou.utils.logger as logger


def check_conan_present():
    out, err = call_process(['conan', '--version'], capture=True, realtime=False)
    if err is not None:
        logger.CriticalError('Conan is missing, unable to proceed')
        return False

    prefix = 'Conan version'
    prefix_pos = out.find(prefix)
    version = out[prefix_pos + len(prefix):].strip()

    logger.Success(f"Conan {version} present")
    return True


def get_conan_settings():
    user_dir = os.path.expanduser('~')
    conan_settings_path = os.path.join(user_dir, '.conan/settings.yml')

    if os.path.exists(conan_settings_path):
        with open(conan_settings_path, encoding='utf-8') as conan_settings_file:
            loader = LineLoader(conan_settings_file.read())
            return loader.get_single_data()

    return None


def call_conan(args : List[str], cwd : str, venv : Optional[VirtualEnvironment] = None):
    verbose = logger.IsVisible(logger.ELogLevel.Verbose)

    out = None
    err = None

    if venv is not None:
        out, err = call_process_venv(['conan'] + args, venv, capture = not verbose, realtime = verbose, cwd = cwd)
    else:
        out, err = call_process(['conan'] + args, capture = not verbose, realtime = verbose, cwd = cwd)

    if err is not None:
        if not verbose:
            logger.CriticalError(f'Conan command failed: {" ".join(args)}')
            logger.Error(out)

        logger.CriticalError(f"Conan returned error '{err}'")
        return False
    return True



from typing import List, Dict
from dataclasses import dataclass
from copy import deepcopy
from zetsubou.project.model.config_string import EDefaultConfigSlots

from zetsubou.utils import logger


@dataclass
class ConfigVariant:
    config_string: str
    slots: Dict[str, str]

    def get_slot(self, slot):
        if isinstance(slot, EDefaultConfigSlots):
            return self.slots[slot.name]
        return self.slots[slot]


def get_config_matrix_os_name(config_string: str):
    return config_string.replace('-', '_')


@dataclass
class ConfigMatrix:
    variants: List[ConfigVariant]
    slots: List[str]

    def has_variant(self, variant_string:str):
        return any(filter(lambda v : v.config_string == variant_string, self.variants))

    # Queries variants matching slot values present in dictionary
    # Egx. { 'configuration' : 'DEBUG' }
    def find_variants(self, data: dict):
        results = []

        for variant in self.variants:
            is_variant_ok = True

            for filter_name, filter_value in data.items():
                if filter_name not in self.slots:
                    raise ValueError(f"Unknown slot '{filter_name}'")

                val = variant.get_slot(filter_name)
                if val != filter_value:
                    is_variant_ok = False

            if is_variant_ok:
                results.append(variant)

        return results


    def get_config_name(self, pattern : str, variant : str):
        if isinstance(pattern, EDefaultConfigSlots):
            pattern = pattern.name

        pattern_slots = pattern.split('-')
        variant_slots = variant.split('-')
        out_values = []

        for slot in pattern_slots:
            if slot not in self.slots:
                raise ValueError(f"No slot '{slot}' in config matrix! Available slots: {self.slots}")
            slot_index = self.slots.index(slot)
            out_values.append(variant_slots[slot_index])

        return '-'.join(out_values)


def parse_config_string(conf: str):
    slot_defs = conf.split('-')
    result = []

    if len(slot_defs) == 0:
        return None

    for slot in slot_defs:
        if slot[0] != '{' or slot[len(slot) - 1] != '}':
            continue
        result.append(slot[1:len(slot)-1])

    return result

def create_config_matrix(config_string:str, domain:list) -> ConfigMatrix:
    slots = parse_config_string(config_string)
    config_variants = []

    template = config_string

    for slot_name, values in domain.items():
        if len(values) == 0:
            logger.CriticalError(f'Unable to build list of values for slot \'{slot_name}\'')
            return None

    for slot_name, _ in domain.items():
        if slot_name not in slots:
            domain.pop(slot_name)

    def RecurseBuildVariants(slot_itr:int=0, field_values={}):
        if slot_itr >= len(domain):
            config_variants.append(ConfigVariant(
                config_string=template.format(**field_values),
                slots=deepcopy(field_values)
            ))
        else:
            key, values = list(domain.items())[slot_itr]
            for val in values:
                field_values[key] = val
                RecurseBuildVariants(slot_itr+1, field_values)

    RecurseBuildVariants()

    return ConfigMatrix(variants=config_variants, slots=slots)



from argparse import ArgumentParser
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.commands.install import Install
from zetsubou.commands.generate import Generate


class Config(Command):
    def install(self):
        return Command.get_command_instance(Install)

    def gen(self):
        return Command.get_command_instance(Generate)

    @property
    def name(self):
        return 'config'

    @property
    def desc(self):
        return f'{self.install().desc} {self.gen().desc}'

    @property
    def help(self):
        return 'Calls [install] and [gen].'

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    @property
    def needs_project_loaded(self):
        return True

    def OnExecute(self, context: CommandContext):
        self.install().OnExecute(context)
        self.gen().OnExecute(context)



# pyright: reportMissingImports=false
# pylint: disable=import-error, bare-except

# This file can be imported from conan generator
# So it needs to support importing via absolute and relative packages
import configparser
from typing import Optional
from conan import ConanFile

try:
    from zetsubou.project.model.configuration_enums import EBaseConfiguration
    from zetsubou.project.model.runtime_library import ERuntimeLibrary
    from zetsubou.project.model.toolchain_enums import ECompilerFamily, ECppStandard
    from zetsubou.project.model.platform_enums import EArch, ESystem, EVersionSelector
    from zetsubou.project.model.toolchain_profile import Profile, IToolchainProfile, Windows_ToolchainProfile, Linux_ToolchainProfile
    from zetsubou.utils.yaml_simple_writer import to_yaml
except:
    pass


try:
    from configuration_enums import EBaseConfiguration
    from runtime_library import ERuntimeLibrary
    from toolchain_enums import ECompilerFamily, ECppStandard
    from platform_enums import EArch, ESystem, EVersionSelector
    from toolchain_profile import Profile, IToolchainProfile, Windows_ToolchainProfile, Linux_ToolchainProfile
    from yaml_simple_writer import to_yaml
except:
    pass


defined = lambda o: o is not None


def os_from_conan(system:str) -> Optional[ESystem]:
    return {
        'Windows' : ESystem.Windows,
        'Linux' : ESystem.Linux
    }.get(system, None)


def arch_from_conan(arch:str) -> Optional[EArch]:
    return {
        'x86' : EArch.x86,
        'x86_64' : EArch.x64
    }.get(arch, None)


def compiler_family_from_conan(compiler:str) -> Optional[ECompilerFamily]:
    return {
        'Visual Studio' : ECompilerFamily.MSVC,
        'msvc' : ECompilerFamily.MSVC,
        'gcc' : ECompilerFamily.GCC,
        'clang' : ECompilerFamily.CLANG,
    }.get(compiler, None)


def runtime_from_conan(runtime:str):
    return {
        'dynamic' : ERuntimeLibrary.dynamic,
        'static' : ERuntimeLibrary.static
    }.get(runtime, None)


def cppstd_from_conan(cppstd:str) -> Optional[ECppStandard]:
    return {
        '11' : ECppStandard.cpp11,
        '14' : ECppStandard.cpp14,
        '17' : ECppStandard.cpp17,
        '20' : ECppStandard.cpp20,
        '23' : ECppStandard.cpp23,
    }.get(cppstd, None)


def build_type_from_conan(build_type:str) -> Optional[EBaseConfiguration]:
    return {
        'Debug' : EBaseConfiguration.DEBUG,
        'Release' : EBaseConfiguration.RELEASE,
        'RelWithDebInfo' : EBaseConfiguration.RELEASE_WITH_DEBUG_INFO
    }.get(build_type, None)


def runtime_library_from_conan(conanfile:ConanFile, build_type:EBaseConfiguration) -> ERuntimeLibrary:
    # always present
    runtime = conanfile.settings.get_safe('compiler.runtime')
    if runtime is None:
        conanfile.output.error('Runtime is not set!')
        return None

    # only in new compilers
    runtime_type = conanfile.settings.get_safe('compiler.runtime_type')

    if runtime_type is not None:
        if runtime_type == 'Release' and build_type not in [EBaseConfiguration.RELEASE, EBaseConfiguration.RELEASE_WITH_DEBUG_INFO]:
            conanfile.output.error('Runtime type doesnt match build type!')
            return None
        elif runtime_type == 'Debug' and build_type != EBaseConfiguration.DEBUG:
            conanfile.output.error('Runtime type doesnt match build type!')
            return None

        # in new compilers only two values are supported, debug flag is managed by build type
        return {
            'dynamic' : ERuntimeLibrary.dynamic,
            'static'  : ERuntimeLibrary.static
        }.get(runtime, None)

    else:
        deduced_runtime, deduced_type = {
            'MD'  : (ERuntimeLibrary.dynamic, EBaseConfiguration.RELEASE),
            'MDd' : (ERuntimeLibrary.dynamic, EBaseConfiguration.DEBUG),
            'MT'  : (ERuntimeLibrary.static, EBaseConfiguration.RELEASE),
            'MTd' : (ERuntimeLibrary.static, EBaseConfiguration.DEBUG),
        }.get(runtime, (None, None))

        if deduced_runtime is None or deduced_type is None:
            conanfile.output.error('Invalid value for runtime!')
            return None

        if deduced_type == EBaseConfiguration.RELEASE and build_type in [EBaseConfiguration.RELEASE, EBaseConfiguration.RELEASE_WITH_DEBUG_INFO]:
            return deduced_runtime

        if deduced_type == EBaseConfiguration.DEBUG and build_type == EBaseConfiguration.DEBUG:
            return deduced_type

    conanfile.output.error('Unable to determine runtime library!')
    return None


def export_profile(conanfile:ConanFile, platform_filename:str) -> bool:
    build_type = build_type_from_conan(conanfile.settings.get_safe('build_type'))
    if build_type is None:
        conanfile.output.error('Unable to setup Profile')
        return None

    compiler = toolchain_from_settings(conanfile, build_type)
    if compiler is None:
        conanfile.output.error('Unable to setup Toolchain')
        return False

    profile = Profile(build_type=build_type, toolchains=[ compiler ])

    with open(platform_filename, 'w') as out_file:
        out_file.write(to_yaml(profile))

    return True


def toolchain_from_settings(conanfile:ConanFile, build_type:EBaseConfiguration) -> Optional[IToolchainProfile]:
    arch_host = None
    arch_target = None
    arch = conanfile.settings.get_safe('arch')
    if arch is None:
        arch_host = conanfile.settings.get_safe('arch_build')
        arch_target = conanfile.settings.get_safe('arch_target')
    else:
        arch_target = arch
        arch_host = arch

    arch_host = arch_from_conan(arch_host)
    arch_target = arch_from_conan(arch_target)
    os_host = os_from_conan(conanfile.settings.get_safe('os'))

    if not defined(os_host) or not (defined(arch_target) and defined(arch_host)):
        return None

    compiler_family = compiler_family_from_conan(conanfile.settings.get_safe('compiler'))
    if compiler_family is None:
        print('Unable to discover compiler family')
        return None

    compiler_version = conanfile.settings.get_safe('compiler.version')
    if compiler_version is None:
        compiler_version = EVersionSelector.latest

    cppstd = cppstd_from_conan(conanfile.settings.get_safe('compiler.cppstd'))
    if cppstd is None:
        cppstd = ECppStandard.default()

    # if compiler_family in (ECompilerFamily.MSVC, ECompilerFamily.CLANG, ECompilerFamily.CLANG_CL):
    if os_host == ESystem.Windows:
        compiler_toolset = conanfile.settings.get_safe('compiler.toolset')
        if compiler_toolset is None:
            compiler_toolset = EVersionSelector.latest

        runtime_library = runtime_library_from_conan(conanfile, build_type)
        if runtime_library is None:
            return None

        return Windows_ToolchainProfile(
            target_arch=arch_target,
            compiler_family=compiler_family,
            compiler_version=compiler_version,
            cppstd=cppstd,
            toolset=compiler_toolset,
            runtime=runtime_library
        )
    else:
        compiler_libcxx = conanfile.settings.get_safe('compiler.libcxx')
        if compiler_libcxx is None:
            print('Unable to discover compiler libcxx')
            return None

        return Linux_ToolchainProfile(
            target_arch=arch_target,
            compiler_family=compiler_family,
            compiler_version=compiler_version,
            cppstd=cppstd,
            libcxx=compiler_libcxx
        )


def toolchain_from_profile(conan_profile:str) -> Optional[IToolchainProfile]:
    config = configparser.ConfigParser()

    with open(conan_profile, 'r') as file:
        config.read_file(file)

        if not config.has_section('settings'):
            return None

        settings = config['settings']

        arch = arch_from_conan(settings.get('arch'))
        if arch is None:
            return None

        system = os_from_conan(settings.get('os'))
        if system is None:
            return None

        compiler_family = compiler_family_from_conan(settings.get('compiler'))
        if compiler_family is None:
            return None

        compiler_version = settings.get('compiler.version')
        if compiler_version is None:
            compiler_version = EVersionSelector.latest

        cppstd = cppstd_from_conan(settings.get('compiler.cppstd'))
        if cppstd is None:
            cppstd = ECppStandard.default()

        if system == ESystem.Windows:
            compiler_runtime = runtime_from_conan(settings.get('compiler.runtime'))
            if compiler_runtime is None:
                return None

            # Used to be compiler.toolset in Visual Studio and Clang
            # Now only Clang defines it (msvc probably guesses latest from version)
            compiler_toolset = settings.get('compiler.runtime_version')
            if compiler_toolset is None:
                compiler_toolset = EVersionSelector.latest

            return Windows_ToolchainProfile(
                target_arch=arch,
                compiler_family=compiler_family,
                compiler_version=compiler_version,
                cppstd=cppstd,
                toolset=compiler_toolset,
                runtime=compiler_runtime
            )

        elif system == ESystem.Linux:
            compiler_libcxx = settings.get('compiler.libcxx')
            if compiler_libcxx is None:
                return None

            return Linux_ToolchainProfile(
                target_arch=arch,
                compiler_family=compiler_family,
                compiler_version=compiler_version,
                cppstd=cppstd,
                libcxx=compiler_libcxx
            )

        return None



import os
import venv
from dataclasses import dataclass
from typing import List, Optional
from fs.base import FS

from zetsubou.project.model.project_template import Project
from zetsubou.utils.common import fix_path
from zetsubou.utils.error import ProjectError
import zetsubou.utils.logger as logger


# Class which contains paths to scripts which activate and deactivate virtual environment
# We cannot assume scripts have a particular name due to conan changing convention mid development
@dataclass
class VirtualEnvironment:
    activate: str
    deactivate: str

    @property
    def path(self):
        return os.path.dirname(self.activate)


# Generates new, clean virtual environment in provided FS
def generate_empty_environment(project_fs: FS):
    env_builder = venv.EnvBuilder()
    venv_fs = project_fs.geturl('build/venv', purpose='fs')
    venv_fs = venv_fs.split('osfs://')[1]
    env_builder.create(venv_fs)
    logger.Info('Created new virtual environment')


# Returns path to file which activates virtual environment
# Since conan is quirky, deactivation file might not exists before activation is first called
# Hence we need to assume path here
def resolve_venv(project_fs : FS, proj : Project, proj_path : str) -> Optional[VirtualEnvironment]:
    if proj.config.venv == '':

        def find_script(patterns : List[str]):
            for path in project_fs.walk.files(filter=patterns):
                env_dir = path
                if env_dir.startswith('/'):
                    env_dir = env_dir[1:]
                return env_dir
            return ''

        activate = fix_path(find_script(['activate.*', 'conanbuild.*', 'conanbuildenv*']))
        deactivate = fix_path(find_script(['deactivate.*', 'deactivate_conanbuild.*', 'deactivate_conanbuildenv*']))

        # Conan activation script generates deactivation on first call
        if activate != '' and deactivate == '':
            deactivate = activate.replace('conanbuildenv', 'deactivate_conanbuildenv')

        if activate == '' or deactivate == '':
            proj_yml_path = proj.get_loaded_from_file()
            logger.CriticalError('Virtual environment not found! \n'
                f" Please, either set path to config.venv in {proj_yml_path}, run 'zetsubou install' command, or place 'activate' and 'deactivate' scripts in one of the project subdirectories.")
            return None

        return VirtualEnvironment(activate=activate, deactivate=deactivate)

    else:
        venv = os.path.abspath(os.path.join(proj_path, proj.config.venv))

        def find_script_os(search_root : str, prefixes : List[str]) -> Optional[str]:
            for _, _, files in os.walk(search_root):
                for file in files:
                    if any(filter(lambda p, f=file : f.find(p) != -1, prefixes)):
                        return file
            return ''

        if os.path.exists(venv) and os.path.isdir(venv):
            activate = fix_path(find_script_os(venv, ['activate.', 'conanbuild.', 'conanbuildenv']))
            deactivate = fix_path(find_script_os(venv, ['deactivate.', 'conanbuild_deactivate.', 'deactivate_conanbuildenv']))

            # Conan activation script generates deactivation on first call
            if activate != '' and deactivate == '':
                deactivate = activate.replace('conanbuildenv', 'deactivate_conanbuildenv')

            if activate == '' or deactivate == '':
                proj_yml_path = proj.get_loaded_from_file()
                logger.CriticalError('Virtual environment not found! \n'
                    f" Path '{proj.config.venv}' doesn't contain 'activate' and 'deactivate' scripts! Run 'zetsubou install' command or create them manually.")
                return None

            return VirtualEnvironment(activate=activate, deactivate=deactivate)

        else:
            raise ProjectError(proj.config.format_field_message("venv", "error",
                f"Virtual environment not found! Path '{proj.config.venv}' doesnt exist or is not a directory!",
                logger.GetErrorFormat())
            )



import os
import fs
from pathlib import Path

from dataclasses import dataclass, is_dataclass, fields
from enum import Enum
from inspect import isclass
from typing import Optional, List, Any, Set, Callable, Tuple, Generic, TypeVar
from typing_inspect import get_origin, is_optional_type
from zetsubou.fastbuild.to_fastbuild import compiler_family_to_fastbuild

from zetsubou.project.config_matrix import ConfigVariant, get_config_matrix_os_name
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.runtime.emit import EDefaultTargets, Emitter, EmitContext
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.project.runtime.resolve import ResolvedTarget, TargetVariant
from zetsubou.project.model.target import Target, TargetReference, Source
from zetsubou.project.model.kind import ETargetKind, is_target_kind_linkable, is_target_library, is_target_prebuild_step
from zetsubou.project.model.toolchain import Toolchain, ToolchainDefinition
from zetsubou.utils.common import join_unique, list_to_string, split, filter_none

import zetsubou.fastbuild.target_kinds as target_kinds
import zetsubou.fastbuild.vcxproject as vcxproject
import zetsubou.fastbuild.vssolution as vssolution
from zetsubou.fastbuild.compiler import Compiler
from zetsubou.system.windows.msvc import arch_to_msvc_platform
from zetsubou.utils.subprocess import call_process_venv


FASTBUILD_BFF_HEADER = '// Generated by Zetsubou, do not manually edit!'
FASTBUILD_ALL_TARGET = 'AllTargets'
PROJ_OUTPUT_DIR = '^$(SolutionDir)/build'
FBUILD_OUTPUT_DIR = 'build/fbuild'
DEFAULT_INDENT = 4
T1 = TypeVar('T1')
T2 = TypeVar('T2')


class bff_writer:
    indent : List[int]
    str_buff : List[str]

    def __init__(self):
        self.indent = [0]
        self.str_buff = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def newline(self):
        self.str_buff.append('\n')
        self.str_buff.append(" " * sum(self.indent))

    def write(self, content : str):
        assert(isinstance(content, str))
        self.str_buff.append(content)

    def write_line(self, content : str):
        self.newline()
        self.write(content)

    def push_indent(self, val : int = DEFAULT_INDENT):
        self.indent.append(val)

    def pop_indent(self, val : int = 0):
        popped = self.indent.pop()
        if val != 0:
            assert(popped == val)

    def push_cursor_indent(self):
        indent_sum = sum(self.indent)
        cursor = self.cursor_pos()
        self.push_indent(cursor - indent_sum)

    def cursor_pos(self):
        buff_len = len(self.str_buff)
        itr = -1
        pos = 0
        while itr != -buff_len:
            for char in self.str_buff[itr]:
                if char == '\n':
                    return pos
                else:
                    pos += 1
            itr -= 1
        return pos

    def to_string(self):
        return ''.join(self.str_buff)

    @staticmethod
    def is_empty(obj) -> bool:
        if not is_optional_type(type(obj)):
            return False
        if bff_writer.is_list(obj):
            return len(obj) == 0
        return obj is None

    @staticmethod
    def is_list(obj) -> bool:
        t = type(obj)
        return t == list or get_origin(t) == list

    @staticmethod
    def is_single_value(obj) -> bool:
        t = type(obj)
        if bff_writer.is_list(obj):
            return False
        if t == TargetReference:
            return True
        if t == dict or is_dataclass(t):
            return False
        if t in [str, int, bool]:
            return True
        if isclass(t) and issubclass(t, Enum):
            return True
        return False

    @staticmethod
    def is_struct(obj) -> bool:
        return not bff_writer.is_single_value(obj)

    @staticmethod
    def get_fields(obj):
        t = type(obj)
        if t is dict:
            return obj.items()
        elif is_dataclass(obj):
            return { f.name : getattr(obj, f.name) for f in fields(obj) }.items()
        else:
            return vars(obj).items()


@dataclass
class bff_using(Generic[T1, T2]):
    name : str
    obj : T1
    ref : T2


@dataclass
class bff_includes:
    includes: List[str]
    system_includes: List[str]


class bff_struct:
    writer : bff_writer
    field_map : Set[str]

    STRUCT_BEGIN = '['
    STRUCT_END = ']'

    def __init__(self, writer : bff_writer, is_keyword : bool = False) -> None:
        self.writer = writer
        self.field_map = set()

        if is_keyword:
            self.STRUCT_BEGIN = '{'
            self.STRUCT_END = '}'

    def __enter__(self):
        self.writer.write_line(self.STRUCT_BEGIN)
        self.writer.push_indent()

        return self

    def next(self):
        self.writer.newline()

    def has_field(self, name : str):
        return name in self.field_map

    def write_field(self, name, value, no_quotations : bool = False):
        self.next()

        if not self.has_field(name):
            self.field_map.add(name)
            write_field(self.writer, name, value, no_quotations)
        else:
            write_append_field(self.writer, name, value, False, no_quotations)

    def write_assign(self, name : str, value : str):
        write_assign(self.writer, name)
        self.writer.write(value)

    def write_usings(self, usings : List[Any]):
        for using in usings:
            # Register names for Using objects, so we append their fields instead of overwritting
            if isinstance(using, bff_using):
                for name, _ in bff_writer.get_fields(using.ref):
                    self.field_map.add(name)
                write_using(self.writer, using.name)
            else:
                write_using(self.writer, using)

    def write_object(self, obj : Any, hidden_fields : List[str] = []):
        for name, value in self.writer.get_fields(obj):
            if not self.writer.is_empty(value) and name not in hidden_fields:
                self.write_field(name, value)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.pop_indent()
        self.writer.write_line(self.STRUCT_END)


class bff_foreach(bff_struct):
    collection = str
    def __init__(self, writer : bff_writer, collection : str):
        super().__init__(writer, is_keyword=True)
        self.collection = collection

    def __enter__(self):
        self.writer.write_line(f"ForEach( .Itr in {self.collection} )")
        super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)


class bff_list:
    writer : bff_writer
    obj : List[Any]
    list_len : int
    list_struct : bool
    list_iter : int

    LIST_BEGIN = '{'
    LIST_END = '}'
    LIST_SEPARATOR = ','

    def __init__(self, writer : bff_writer, obj : List[Any]) -> None:
        self.writer = writer
        self.obj = obj
        self.list_len = len(obj)
        self.list_iter = 0
        self.list_struct = False if self.list_len == 0 else writer.is_struct(obj[0])

    def __enter__(self):
        self.writer.write(bff_list.LIST_BEGIN)
        if self.list_struct:
            self.writer.push_indent()
        else:
            self.writer.write(' ')
            self.writer.push_cursor_indent()
        return self

    def next(self):
        if self.list_iter != 0:
            self.writer.write(bff_list.LIST_SEPARATOR)
            if not self.list_struct:
                self.writer.newline()
        self.list_iter += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.pop_indent()
        if self.list_struct:
            self.writer.write_line(bff_list.LIST_END)
        else:
            self.writer.write(' ')
            self.writer.write(bff_list.LIST_END)


class bff_indent:
    writer : bff_writer
    indent : int

    def __init__(self, *, writer : bff_writer, indent : int = DEFAULT_INDENT) -> None:
        self.writer = writer
        self.indent = indent

    def __enter__(self):
        self.writer.push_indent(self.indent)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.pop_indent(self.indent)


def write_value(writer : bff_writer, value, no_quotations : bool = False):
    if writer.is_empty(value):
        return

    if type(value) in [float, int, bool]:
        writer.write(str(value).lower())
        return

    serialized : Optional[str] = to_str(value)
    if serialized is not None:
        if no_quotations:
            writer.write(serialized)
        else:
            if serialized.find('"') != -1:
                writer.write(f"'{serialized}'")
            else:
                writer.write(f'"{serialized}"')
        return

    if writer.is_list(value):
        with bff_list(writer, value) as array:
            for element in value:
                if not writer.is_empty(element):
                    array.next()
                    write_value(writer, element, no_quotations=no_quotations)
        return

    elif writer.is_struct(value):
        with bff_struct(writer) as struct:
            for name, prop in writer.get_fields(value):
                if not writer.is_empty(prop):
                    struct.next()
                    write_field(writer, name, prop, no_quotations=no_quotations)
        return

    writer.write(f'"NOT_IMPLEMENTED: {type(value)}"')


def write_comment_header(writer : bff_writer, text : str):
    separator = '/////////////////////////////////////////////////////////'
    writer.write_line(separator)
    writer.write_line(f'// {text}')
    writer.write_line(separator)


def write_assign(writer : bff_writer, name : str):
    writer.write(f'.{name} = ')


def write_field(writer : bff_writer, name : str, value, no_quotations : bool = False):
    write_assign(writer, name)
    write_value(writer, value, no_quotations=no_quotations)


def write_append_field(writer : bff_writer, name : str, value, outer_scope:bool=False, no_quotations : bool = False):
    writer.write(f'^{name} + ' if outer_scope else f'.{name} + ')
    write_value(writer, value, no_quotations=no_quotations)


def write_include(writer : bff_writer, include):
    writer.write_line(f'#include "{include}"')


def write_using(writer : bff_writer, using):
    writer.write_line(f'Using( .{using} )')


def write_includes(writer : bff_writer, includes : List[Any], relative_to : str = ''):
    if includes is not None:
        for include in includes:
            if relative_to != '':
                include = fs.path.relativefrom(relative_to, include)
            write_include(writer, include)


def write_keyword(writer : bff_writer, keyword : str, name_param : str):
    writer.write_line(f"{keyword}( '{name_param}' )")


def write_object(writer : bff_writer, obj, *, name : str = '', local_usings : List[str] = []):
    if name == '' and hasattr(obj, 'Alias'):
        name = getattr(obj, 'Alias')

    writer.newline()
    write_keyword(writer, obj.keyword(), name)

    with bff_struct(writer, is_keyword=True) as struct:
        struct.write_usings(local_usings)
        struct.write_object(obj)

    writer.newline()


##############################################################################################################


def get_linker_output_name(target: ResolvedTarget):
    # support custom!
    if target.target.config.kind == ETargetKind.DYNAMIC_LIBRARY:
        return f'{target.target.target}.dll'
    if target.target.config.kind == ETargetKind.EXECUTABLE:
        return f'{target.target.target}.exe'


def create_bff_writer():
    writer = bff_writer()
    writer.write(FASTBUILD_BFF_HEADER)
    return writer


def get_min_max_msvc_version(context : EmitContext, proj : ProjectTemplate) -> Tuple[str, str]:
    vmin = None
    vmax = None

    for plat in proj.platforms:
        if plat.host_system == context.host_system and plat.target_arch == context.host_arch:
            for tool in proj.platform_toolchains[plat.platform]:
                if vmin is None or tool.definition.ide_version < vmin:
                    vmin = tool.definition.ide_version
                if vmax is None or tool.definition.ide_version > vmax:
                    vmax = tool.definition.ide_version

    return (str(vmin), str(vmax))


def find_cli_tool(tool_name : str, proj : ProjectTemplate):
    tool = proj.find_cli_tool(tool_name)
    return tool


def find_toolchain(plat_name : str, tool_name : str, proj : ProjectTemplate) -> bff_using[Toolchain, ToolchainDefinition]:
    tool = proj.find_toolchain(plat_name, tool_name)
    if tool is not None:
        return bff_using(name=tool.name, obj=tool, ref=tool.definition)
    return None


def find_configuration(name : str, proj : ProjectTemplate) -> Configuration:
    return proj.find_config(name)


def find_base_platform_name(name : str, proj : ProjectTemplate) -> str:
    platform = proj.find_platform(name)
    return arch_to_msvc_platform(platform.target_arch)


def filter_target_linkable(ref : TargetReference):
    return is_target_kind_linkable(ref.target.config.kind)


def filter_target_libraries(ref : Target):
    return is_target_library(ref.config.kind)


def filter_target_executable(ref : Target):
    return not is_target_library(ref.config.kind)


def filter_target_prebuild_step(ref : Target):
    return is_target_prebuild_step(ref.config.kind)


def to_str(obj) -> Optional[str]:
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, TargetReference) or issubclass(type(obj), Enum):
        return obj.name
    else:
        return None


def get_upward_project_path(root : str, path : str):
    depth = root.count('/') + 2
    return f'{depth * "../"}{path}'


class FastbuildEmitter(Emitter):
    subprojects = []
    main_file_name = ''


    def __init__(self):
        self.main_file_name = self.get_main_bff_path()


    def project_to_output_path(self, context : EmitContext, path : str, project_subdir : str = ''):
        abs_dir_path = path if os.path.isabs(path) else fs.path.join(context.fs_root, path)
        proj_path = fs.path.join(context.ROOT_DIR, project_subdir)
        abs_out_path = fs.path.join(context.fs_root, proj_path)
        result = fs.path.relativefrom(abs_out_path, abs_dir_path)
        return result


    def get_target_output_path(self, context : EmitContext, target: ResolvedTarget):
        return self.project_to_output_path(context, fs.path.dirname(target.target.get_loaded_from_file()))


    def get_bin_output_path(self, config_name : str):
        return f'{PROJ_OUTPUT_DIR}/obj/{config_name}'


    def get_compiler_output_path(self, config_str : str, subdir : str):
        return f'build/{subdir}/{config_str}'


    @staticmethod
    def get_copy_distributed_files_target_name(config_variant_name:str):
        return f'COPY_DistributedFiles_{config_variant_name}'


    @staticmethod
    def get_runner_path():
        return 'cd ^$(SolutionDir) &amp; .\\build\\scripts\\runner.bat'


    @staticmethod
    def get_vcxproj_path(name : str):
        return f'build/projects/{name}.vcxproj'


    @staticmethod
    def get_main_bff_path():
        return f'{FBUILD_OUTPUT_DIR}/fbuild.bff'


    @staticmethod
    def get_fbuild_build_command(context: EmitContext, target_name : str):
        cmd = f'fbuild.exe -config {FastbuildEmitter.get_main_bff_path()} -ide -wrapper -cache {target_name}'
        if context.project_template.project.config.verbose_build:
            cmd += ' -verbose'
        return cmd


    @staticmethod
    def get_zetsubou_build_command(context: EmitContext, target_name: str):
        return f'zetsubou build {context.project_file} --ide --nologo --target-variant {target_name}'


    def get_build_command(self, context: EmitContext, target_name : str):
        return f'{FastbuildEmitter.get_runner_path()} {FastbuildEmitter.get_zetsubou_build_command(context, target_name)}'


    @staticmethod
    def get_fbuild_rebuild_command(context: EmitContext, target_name : str):
        cmd = f'fbuild.exe -config {FastbuildEmitter.get_main_bff_path()} -ide -wrapper -cache -clean {target_name}'
        if context.project_template.project.config.verbose_build:
            cmd += ' -verbose'
        return cmd


    def get_rebuild_command(self, context:EmitContext, target_name:str):
        return f'{FastbuildEmitter.get_runner_path()} {FastbuildEmitter.get_fbuild_rebuild_command(context, target_name)}'


    @staticmethod
    def get_zetsubout_update_build_command(context: EmitContext):
        return f'zetsubou gen {context.project_file} --ide --nologo'


    def get_update_build_command(self, context: EmitContext):
        return f'{FastbuildEmitter.get_runner_path()} {FastbuildEmitter.get_zetsubout_update_build_command(context)}'


    @staticmethod
    def get_zetsubout_update_rebuild_command(context: EmitContext):
        return f'zetsubou regen {context.project_file} --ide --nologo'


    # Formats path containing {root} and {config_variant} like so:
    # {root}/mydir -> C:\Projects\MyProject\mydir
    # build/generated/{config_variant} -> build/generated/Windows_Debug_MSVC_x64_v193_v143
    @staticmethod
    def format_path(format_path:str, context:EmitContext, config_variant:ConfigVariant):
        return format_path.format(root=context.fs_root, config_variant=get_config_matrix_os_name(config_variant.config_string))


    def get_update_rebuild_command(self, context: EmitContext):
        return f'{FastbuildEmitter.get_runner_path()} {FastbuildEmitter.get_zetsubout_update_rebuild_command(context)}'


    def emit_runner_script(self, context: EmitContext):
        relative_activate = self.project_to_output_path(context, context.fs_venv.activate).replace('/', '\\')
        relative_deactivate = self.project_to_output_path(context, context.fs_venv.deactivate).replace('/', '\\')
        script = ''

        script += f'CALL {relative_activate}\n'
        script +=  'CALL %* \n'
        script +=  'SET taskexitcode=%errorlevel%\n'
        script += f'CALL {relative_deactivate}\n'
        script +=  'if %taskexitcode% NEQ 0 (\n'
        script +=  '    exit /b %taskexitcode%\n'
        script +=  ')\n'

        context.write_file('build/scripts/runner.bat', script)


    def fill_base_project(self, context : EmitContext, fbuild: target_kinds.LinkableTarget, target: ResolvedTarget, target_variant: TargetVariant, toolchain : Toolchain):
        subdir = 'bin' if target.get_kind() == ETargetKind.EXECUTABLE else 'lib'
        lib_dir = self.get_compiler_output_path(target_variant.config_variant.config_string, subdir)
        fbuild.LinkerOutput = f'{lib_dir}/{get_linker_output_name(target)}'

        fbuild.LinkerOptions = list_to_string(join_unique([
            target_variant.linker_flags.public,
            target_variant.linker_flags.private
        ]))

        # Append library paths as flags with /LIBPATH:
        lib_paths = join_unique([
            [ lib_dir ],
            toolchain.definition.LinkerPaths,
            target_variant.linker_paths.public,
            target_variant.linker_paths.private
        ])

        link_libs_with_refs = join_unique([
            target_variant.link_libraries.public,
            target_variant.link_libraries.private
        ])

        def filter_target_ref(ref):
            return isinstance(ref, TargetReference)

        refs, libs = split(filter_target_ref, link_libs_with_refs)

        # Linking only to matching variant of other targets
        def map_target_ref_to_str(ref):
            if isinstance(ref, TargetReference):
                return ProjectTemplate.compile_target_variant_name(ref.target.config.kind, ref.name, target_variant.config_variant.config_string)
            return ref

        fbuild.Libraries = list(map(map_target_ref_to_str, filter(filter_target_linkable, refs)))
        fbuild.LinkerOptions += list_to_string( [ toolchain.definition.i_linker.link(l) for l in libs ])
        fbuild.LinkerOptions += list_to_string( [ toolchain.definition.i_linker.dir(l)  for l in lib_paths ])

        # Depend on copying distributed files to bin
        fbuild.PreBuildDependencies = [ FastbuildEmitter.get_copy_distributed_files_target_name(target_variant.config_variant.config_string) ]


    def compiler_relative_path(self, context : EmitContext, path : str, target : ResolvedTarget, target_variant : TargetVariant):
        formatted_path = self.format_path(path, context, target_variant.config_variant)
        if not os.path.isabs(formatted_path):
            return fs.path.join( fs.path.dirname(target.target.get_loaded_from_file()), formatted_path)
        else:
            return formatted_path


    def fill_object(self, context : EmitContext, fbuild: target_kinds.ObjTarget, target: ResolvedTarget, target_variant: TargetVariant, root_dir : str, toolchain : Toolchain, source_path : str, includes : bff_includes):
        fbuild.CompilerOutputPath = self.get_compiler_output_path(target_variant.config_variant.config_string, 'obj')
        fbuild.CompilerOutputPrefix = f'{target.target.target}_'

        fbuild.CompilerOptions = list_to_string(join_unique([
            target_variant.compiler_flags.public,
            target_variant.compiler_flags.private
        ]))

        defines = join_unique([
            target_variant.defines.public,
            target_variant.defines.private
        ])

        fbuild.CompilerOptions += list_to_string([ toolchain.definition.i_compiler.cppstd(toolchain.profile.cppstd) ])
        fbuild.CompilerOptions += list_to_string([ toolchain.definition.i_compiler.define(d) for d in defines ])

        fbuild.CompilerOptions += list_to_string([ toolchain.definition.i_compiler.include(i) for i in includes.includes ])
        fbuild.CompilerOptions += list_to_string([ toolchain.definition.i_compiler.system_include(i) for i in includes.system_includes ])

        out_path = self.compiler_relative_path(context, source_path, target, target_variant)

        excluded_paths = list(map(lambda p: self.compiler_relative_path(context, p, target, target_variant), target_variant.source_exclude.paths))

        fbuild.CompilerInputPath = out_path
        fbuild.CompilerInputPattern = target_variant.source.patterns
        fbuild.CompilerInputExcludePath = excluded_paths
        fbuild.CompilerInputExcludePattern = target_variant.source_exclude.patterns


    def fill_library(self, context : EmitContext, fbuild : target_kinds.StaticLibTarget, target: ResolvedTarget, target_variant: TargetVariant, root_dir : str, toolchain : Toolchain):
        # self.fill_object(context, fbuild, target, target_variant, root_dir, toolchain)
        fbuild.CompilerOutputPath = self.get_compiler_output_path(target_variant.config_variant.config_string, 'obj')

        fbuild.LibrarianOptions = list_to_string(join_unique([
            target_variant.librarian_flags.public,
            target_variant.librarian_flags.private
        ]))

        fbuild.LibrarianOutput = f'{self.get_compiler_output_path(target_variant.config_variant.config_string, "lib")}/{target.target.target}.lib'


    def emit_vcxproj(self, writer : bff_writer, vcxproj : vcxproject.VCXProject, project_configs_list : str, includes : List[str]):
        # Conflig list predefine
        writer.newline()
        write_field(writer, project_configs_list, [])

        # Variant includes
        writer.newline()
        write_includes(writer, includes)

        # VCXProject
        writer.newline()
        write_keyword(writer, vcxproject.VCXProject.keyword(), vcxproj.Alias)

        with bff_struct(writer, is_keyword=True) as struct:
            struct.write_object(vcxproj, hidden_fields=['Alias','ProjectConfigs'])
            struct.next()
            write_assign(writer, 'ProjectConfigs')
            writer.write(f'.{project_configs_list}')


    def emit_imported_target(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str) -> List[str]:
        includes = []

        imported_lib_name = f'configs/{target_name}_config.bff'
        write_path = f'{FBUILD_OUTPUT_DIR}/{target_name}/{imported_lib_name}'
        includes.append(imported_lib_name)

        with create_bff_writer() as writer:
            for config_variant in context.config_matrix.variants:
                target_variant = target.variants[config_variant.config_string]

                writer.newline()
                alias = ProjectTemplate.compile_target_variant_name(target.get_kind(), target_name, config_variant.config_string)

                # Target config tuple to be imported into VCXProject
                config_list_entry = f'{get_config_matrix_os_name(alias)}_Config'

                writer.newline()
                write_assign(writer, config_list_entry)

                with bff_struct(writer) as struct:
                    custom_platform = context.config_matrix.get_config_name(EDefaultConfigSlots.platform, target_variant.config_variant.config_string)
                    vs_platform = find_base_platform_name(custom_platform, context.project_template)

                    struct.write_field('Platform', vs_platform)
                    struct.write_field('Config', target_variant.config_variant.config_string) # context.config_matrix.get_config_name('platform-configuration-toolchain', target_variant.config_variant.config_string))

                writer.newline()
                writer.write(f'.{project_configs_list} + {{ .{config_list_entry} }}')

        bff_content = writer.to_string()
        context.write_file(write_path, bff_content)
        return includes


    def emit_custom_build_step(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str) -> List[str]:
        includes = []

        for config_variant in context.config_matrix.variants:
            target_variant = target.variants[config_variant.config_string]
            target_variant_name = ProjectTemplate.compile_target_variant_name(ETargetKind.BUILD_STEP, target_name, config_variant.config_string)
            fb_target_variant_name = get_config_matrix_os_name(target_variant_name)

            cli_tool = find_cli_tool(target.target.config.compiler, context.project_template)
            cli_fullpath = cli_tool.executable_fullpath

            exclude_src_paths = list(map(lambda p: self.compiler_relative_path(context, p, target, target_variant), target_variant.source_exclude.paths))
            output_dir = self.compiler_relative_path(context, cli_tool.output_dir, target, target_variant)

            sub_parts = f'{fb_target_variant_name}_parts'

            with create_bff_writer() as writer:
                writer.newline()
                write_field(writer, sub_parts, [])

                # For each source path
                for idx, source_path in enumerate(target_variant.source.paths):
                    part_name = f'{fb_target_variant_name}_PART_{idx}'

                    writer.newline()
                    write_keyword(writer, 'Exec', part_name)

                    with bff_struct(writer, is_keyword=True) as struct:
                        struct.write_field('ExecExecutable', cli_fullpath)
                        struct.write_field('ExecInputPath', self.compiler_relative_path(context, source_path, target, target_variant))
                        struct.write_field('ExecInputPattern', target_variant.source.patterns)
                        struct.write_field('ExecInputExcludePath', exclude_src_paths)
                        struct.write_field('ExecInputExcludePattern', target_variant.source_exclude.patterns)
                        struct.write_field('ExecArguments', list_to_string(cli_tool.options).replace('%3', output_dir))
                        struct.write_field('ExecOutput', os.path.join(output_dir, cli_tool.output_file))

                    writer.newline()
                    writer.write(f'.{sub_parts} + {{ "{part_name}" }}')
                    writer.newline()

            write_keyword(writer, 'Alias', target_variant_name)
            with bff_struct(writer, is_keyword=True) as struct:
                writer.newline()
                struct.write_assign('Targets', f'.{sub_parts}')

            writer.newline()
            writer.newline()

            custom_platform = context.config_matrix.get_config_name(EDefaultConfigSlots.platform, target_variant.config_variant.config_string)

            proj_config = vcxproject.ProjectConfig()
            proj_config.Platform = find_base_platform_name(custom_platform, context.project_template)
            proj_config.Config = target_variant.config_variant.config_string
            proj_config.Target = target_variant_name
            proj_config.ProjectBuildCommand = self.get_build_command(context, target_variant_name)

            config_name = f'{fb_target_variant_name}_Config'
            write_assign(writer, config_name)
            with bff_struct(writer) as struct:
                struct.write_object(proj_config)

            writer.newline()
            writer.write(f'.{project_configs_list} + {{ .{config_name} }}')

            custom_step_path = f'configs/{fb_target_variant_name}.bff'
            write_path = f'{FBUILD_OUTPUT_DIR}/{target_name}/{custom_step_path}'
            includes.append(custom_step_path)

            bff_content = writer.to_string()
            context.write_file(write_path, bff_content)

        return includes


    def gather_includes(self, context : EmitContext, target : ResolvedTarget, target_variant : TargetVariant, toolchain : Toolchain) -> bff_includes:
        raw_includes = join_unique([
            [".\\"],
            target_variant.includes.public,
            target_variant.includes.private
        ])

        raw_includes_sys = join_unique([
            target_variant.system_includes.public,
            target_variant.system_includes.private,
            toolchain.definition.IncludeDirectories,
        ])

        def preprocess_includes(raw : List[str]) -> List[str]:
            return list(map(lambda p: self.compiler_relative_path(context, p, target, target_variant), raw))

        return bff_includes(
            includes=preprocess_includes(raw_includes),
            system_includes=preprocess_includes(raw_includes_sys)
        )


    def emit_static_library(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str) -> List[str]:
        includes = []

        for config_variant in context.config_matrix.variants:
            target_variant = target.variants[config_variant.config_string]
            toolchain = find_toolchain(target_variant.platform, target_variant.compiler, context.project_template)

            obj_list : List[target_kinds.ObjTarget] = []
            intelisense_includes : List[str] = []

            for idx, source_path in enumerate(target_variant.source.paths):

                headers = self.gather_includes(context, target, target_variant, toolchain.obj)
                intelisense_includes += headers.includes
                intelisense_includes += headers.system_includes

                obj_lib = target_kinds.ObjTarget()
                obj_lib.Alias = ProjectTemplate.compile_target_variant_name(ETargetKind.OBJECT_LIBRARY, target_name, config_variant.config_string)
                obj_lib.Alias += f'_{idx}'
                self.fill_object(context, obj_lib, target, target_variant, context.ROOT_DIR, toolchain.obj, source_path, headers)
                obj_list.append(obj_lib)

            static_lib = target_kinds.StaticLibTarget()
            static_lib.Alias = ProjectTemplate.compile_target_variant_name(ETargetKind.STATIC_LIBRARY, target_name, config_variant.config_string)
            self.fill_library(context, static_lib, target, target_variant, context.ROOT_DIR, toolchain.obj)
            static_lib.LibrarianAdditionalInputs = [ obj.Alias for obj in obj_list ]

            static_lib_filename = get_config_matrix_os_name(static_lib.Alias)
            static_lib_path = f'configs/{static_lib_filename}.bff'
            write_path = f'{FBUILD_OUTPUT_DIR}/{target_name}/{static_lib_path}'
            includes.append(static_lib_path)

            with create_bff_writer() as writer:
                tool_relative = fs.path.relativefrom(fs.path.dirname(write_path), f'{FBUILD_OUTPUT_DIR}/toolchains/toolchain_{toolchain.obj.name}.bff')

                write_include(writer, tool_relative)
                writer.newline()

                self.emit_buildable_target_variant(context, writer, static_lib, obj_list, target_variant, project_configs_list, intelisense_includes)

                bff_content = writer.to_string()
                context.write_file(write_path, bff_content)

        return includes


    def emit_executable(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str) -> List[str]:
        return self.emit_linkable_target(context, target, target_name, project_configs_list, target_kinds.AppTarget)


    def emit_dynamic_library(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str) -> List[str]:
        return self.emit_linkable_target(context, target, target_name, project_configs_list, target_kinds.DynLibTarget)


    def emit_linkable_target(self, context: EmitContext, target: ResolvedTarget, target_name : str, project_configs_list : str, ctor : Callable[[], target_kinds.LinkableTarget]) -> List[str]:
        includes = []

        for config_variant in context.config_matrix.variants:
            target_variant = target.variants[config_variant.config_string]
            toolchain = find_toolchain(target_variant.platform, target_variant.compiler, context.project_template)

            obj_list : List[target_kinds.ObjTarget] = []
            intelisense_includes : List[str] = []

            for idx, source_path in enumerate(target_variant.source.paths):

                headers = self.gather_includes(context, target, target_variant, toolchain.obj)
                intelisense_includes += headers.includes
                intelisense_includes += headers.system_includes

                obj_lib = target_kinds.ObjTarget()
                obj_lib.Alias = ProjectTemplate.compile_target_variant_name(ETargetKind.OBJECT_LIBRARY, target_name, config_variant.config_string)
                obj_lib.Alias += f'_{idx}'

                self.fill_object(context, obj_lib, target, target_variant, context.ROOT_DIR, toolchain.obj, source_path, headers)
                obj_list.append(obj_lib)

            lnk_target = ctor()
            lnk_target.Alias = ProjectTemplate.compile_target_variant_name(target.get_kind(), target_name, config_variant.config_string)
            self.fill_base_project(context, lnk_target, target, target_variant, toolchain.obj)
            lnk_target.Libraries += [ obj.Alias for obj in obj_list ]

            lnk_target_filename = get_config_matrix_os_name(lnk_target.Alias)
            lnk_target_path = f'configs/{lnk_target_filename}.bff'
            write_path = f'{FBUILD_OUTPUT_DIR}/{target_name}/{lnk_target_path}'

            includes.append(lnk_target_path)

            with create_bff_writer() as writer:
                tool_relative = fs.path.relativefrom(fs.path.dirname(write_path), f'{FBUILD_OUTPUT_DIR}/toolchains/toolchain_{toolchain.obj.name}.bff')

                write_include(writer, tool_relative)
                writer.newline()

                self.emit_buildable_target_variant(context, writer, lnk_target, obj_list, target_variant, project_configs_list, intelisense_includes)

                bff_content = writer.to_string()
                context.write_file(write_path, bff_content)

        return includes


    def emit_buildable_target_variant(self, context: EmitContext, writer : bff_writer, out_target : target_kinds.BaseTarget,
        obj_list : List[target_kinds.BaseTarget], target_variant : TargetVariant, append_to_list : str, includes : List[str]):

        toolchain = find_toolchain(target_variant.platform, target_variant.compiler, context.project_template)

        # Sub objects, per source
        for obj in obj_list:
            write_keyword(writer, obj.keyword(), obj.Alias)
            with bff_struct(writer, is_keyword=True) as struct:
                struct.write_usings([ toolchain ])
                struct.write_object(obj)
            writer.newline()

        # Target definition
        write_keyword(writer, out_target.keyword(), out_target.Alias)
        with bff_struct(writer, is_keyword=True) as struct:
            struct.write_usings([ toolchain ])
            struct.write_object(out_target)

        writer.newline()

        # Target config tuple to be imported into VCXProject
        config_list_entry = f'{get_config_matrix_os_name(out_target.Alias)}_Config'
        write_assign(writer, config_list_entry)

        with bff_struct(writer) as struct:
            custom_platform = context.config_matrix.get_config_name(EDefaultConfigSlots.platform, target_variant.config_variant.config_string)
            vs_platform = find_base_platform_name(custom_platform, context.project_template)

            def relative_include_to_solution(include_path : str):
                if os.path.isabs(include_path):
                    return include_path
                return f'^$(SolutionDir)\\{include_path}'

            includes = list(map(relative_include_to_solution, includes))

            proj_config = vcxproject.ProjectConfig()
            proj_config.Platform = vs_platform
            proj_config.Config = target_variant.config_variant.config_string
            proj_config.Target = out_target.Alias
            proj_config.PlatformToolset = toolchain.obj.toolset
            proj_config.ProjectBuildCommand = self.get_build_command(context, out_target.Alias)
            proj_config.ProjectRebuildCommand = self.get_rebuild_command(context, out_target.Alias)
            proj_config.OutputDirectory = self.get_bin_output_path(target_variant.config_variant.config_string)
            proj_config.IncludeSearchPath = ';'.join(includes).replace('/', '\\')

            struct.write_object(proj_config)

        writer.newline()
        writer.write(f'.{append_to_list} + {{ .{config_list_entry} }}')


    def emit_toolchain(self, context: EmitContext, toolchain: Toolchain):
        writer = create_bff_writer()
        writer.write_line('#once')

        compiler_family = compiler_family_to_fastbuild(toolchain.definition.CompilerFamily)
        compiler = Compiler(
            Executable=toolchain.definition.Compiler,
            CompilerFamily=compiler_family
        )
        write_object(writer, compiler, name=toolchain.name)

        writer.newline()

        # Emit struct which will be then used by compilation targtes
        write_assign(writer, toolchain.name)

        with bff_struct(writer) as struct:
            struct.write_object(toolchain.definition, hidden_fields=['Compiler', 'CompilerFamily'] + toolchain.definition.get_hidden_fields())
            struct.write_field('Compiler', toolchain.name)

        path = f'{FBUILD_OUTPUT_DIR}/toolchains/toolchain_{toolchain.name}.bff'
        context.write_file(path, writer.to_string())
        return path


    def emit_solution(self, context: EmitContext):
        with create_bff_writer() as writer:
            #
            # Files to include as global project
            #
            project_files = []
            def append_proj_files(element):
                if element is not None:
                    project_files.append(element)

            if context.conan is not None and context.project_template.project.conan is not None:
                append_proj_files(context.project_template.project.conan.build_tools)
                append_proj_files(context.project_template.project.conan.dependencies)
            append_proj_files(context.project_template.project.get_loaded_from_file())

            #
            # Environment
            #
            writer.write_line('#import SystemRoot')
            writer.newline()
            writer.write_line('Settings')
            with bff_struct(writer, is_keyword=True) as struct:
                struct.write_field('Environment', [
                    'SystemRoot=$SystemRoot$',
                    'TMP=$SystemRoot$\\temp',
                ])

            writer.newline()

            project_name = context.project_template.project.project
            sln = vssolution.VSSolution()
            sln.Alias = project_name

            #
            # Toolchains
            #
            for plat in context.project_template.platforms:
                if plat.host_system == context.host_system and plat.host_arch == context.host_arch:
                    for tool in context.project_template.platform_toolchains[plat.platform]:
                        tool_inc = self.emit_toolchain(context, tool)
                        if tool_inc != '':
                            write_include(writer, fs.path.relativefrom(FBUILD_OUTPUT_DIR, tool_inc))

            writer.newline()

            #
            # Copy imported files to bin
            #
            write_comment_header(writer, 'Copy imported files to bin')

            def path_join_if_not_abs(path:str, fs_root:str):
                if os.path.isabs(path):
                    return path
                return os.path.join(fs_root, path)

            def grep_files(file_src:Source, fs_root:str = ''):
                results : List[str] = []
                paths : List[str] = []
                if fs_root != '':
                    paths = list(map(lambda p : path_join_if_not_abs(p, fs_root), file_src.paths))
                else:
                    paths = file_src.paths

                for search_path in paths:
                    for pattern in file_src.patterns:
                        results = join_unique([results, Path(search_path).glob(pattern)])
                return results

            for config_variant in context.config_matrix.variants:
                writer.newline()

                # Gather files per configuration
                dist_files = []

                resolved_target : ResolvedTarget
                for resolved_target in join_unique([context.resolved_targets, context.conan.resolved_targets]):
                    target_fs_root = os.path.dirname(resolved_target.target.get_loaded_from_file())
                    target_variant = resolved_target.variants[config_variant.config_string]
                    found_files = grep_files(target_variant.distribute_files, target_fs_root)
                    dist_files = join_unique([dist_files, list(found_files)])

                write_keyword(writer, 'Copy', FastbuildEmitter.get_copy_distributed_files_target_name(config_variant.config_string))
                with bff_struct(writer, is_keyword=True) as struct:
                    struct.write_field('Source', [str(fullpath) for fullpath in dist_files])
                    # Slash at the end is very important! It indicates that we are copying to folder, not to a file
                    struct.write_field('Dest', f"{self.get_compiler_output_path(config_variant.config_string, 'bin')}/")

            #
            # Config matrix
            #
            writer.newline()
            write_comment_header(writer, 'Config Matrix')

            # Include Targets bff's
            for include in self.subprojects:
                write_include(writer, fs.path.relativefrom(FBUILD_OUTPUT_DIR, include))

            writer.newline()
            for config_variant in context.config_matrix.variants:
                writer.newline()
                variant_name = get_config_matrix_os_name(config_variant.config_string)
                write_assign(writer, variant_name)
                with bff_struct(writer) as struct:
                    custom_platform = context.config_matrix.get_config_name(EDefaultConfigSlots.platform, config_variant.config_string)
                    vs_platform = find_base_platform_name(custom_platform, context.project_template)

                    struct.write_field('Config', config_variant.config_string)
                    struct.write_field('Platform', vs_platform)
                    struct.write_field('SolutionConfig', context.config_matrix.get_config_name(
                        EDefaultConfigSlots.configuration, config_variant.config_string))
                    struct.write_field('SolutionPlatform', context.config_matrix.get_config_name(
                        f'{EDefaultConfigSlots.platform.name}-{EDefaultConfigSlots.toolchain.name}', config_variant.config_string))

            writer.newline()
            writer.newline()
            write_field(writer, 'ConfigMatrix',
                [ f'.{get_config_matrix_os_name(config_variant.config_string)}' for config_variant in context.config_matrix.variants ],
                no_quotations=True)

            #
            # All alias
            #
            all_target_proj = f'{EDefaultTargets.All.name}-proj'

            writer.newline()
            write_comment_header(writer, 'All Alias')
            all_projects = [ f'{target.name}-proj' for target in context.resolved_targets ]

            writer.newline()
            write_field(writer, FASTBUILD_ALL_TARGET, all_projects)

            writer.newline()
            write_keyword(writer, 'Alias', EDefaultTargets.All.name)
            with bff_struct(writer, is_keyword=True) as struct:
                struct.next()
                write_assign(writer, 'Targets')
                writer.write(f'.{FASTBUILD_ALL_TARGET}')

            # Emit separate All aliases per config
            writer.newline()
            for config in context.config_matrix.variants:
                this_all_name = f'{EDefaultTargets.All.name}-{config.config_string}'
                write_keyword(writer, 'Alias', this_all_name)
                with bff_struct(writer, is_keyword=True) as struct:
                    compilable_targets = filter(filter_target_linkable, context.resolved_targets)
                    struct.write_field('Targets', [
                        ProjectTemplate.compile_target_variant_name(t.get_kind(), t.target.target, config.config_string) for t in compilable_targets
                    ])

            writer.newline()
            write_comment_header(writer, 'Solution')
            writer.newline()

            write_keyword(writer, vcxproject.VCXProject.keyword(), all_target_proj)
            with bff_struct(writer, is_keyword=True) as struct:
                struct.write_field('ProjectOutput', self.get_vcxproj_path(EDefaultTargets.All.name))
                struct.write_field('ProjectInputPaths', [ FBUILD_OUTPUT_DIR ])
                struct.write_field('ProjectBasePath', 'build')
                struct.write_field('ProjectFiles', project_files)
                struct.write_field('OutputDirectory', PROJ_OUTPUT_DIR)

                struct.next()
                struct.write_field('Configs', [])

                with bff_foreach(writer, '.ConfigMatrix'):
                    writer.newline()
                    write_assign(writer, 'ThisConfig')

                    with bff_struct(writer) as struct:
                        write_using(writer, 'Itr')
                        all_config_name = f'{EDefaultTargets.All.name}-$Config$'
                        struct.write_field('Target', all_config_name)
                        struct.write_field('ProjectBuildCommand', self.get_build_command(context, all_config_name))
                        struct.write_field('ProjectRebuildCommand', self.get_rebuild_command(context, all_config_name))

                    writer.newline()
                    write_append_field(writer, 'Configs', '.ThisConfig', outer_scope=True, no_quotations=True)

                writer.newline()
                struct.write_assign('ProjectConfigs', '.Configs')

            #
            # Update solution
            #
            update_target_proj = f'{EDefaultTargets.Update.name}-proj'

            writer.newline()
            write_keyword(writer, vcxproject.VCXProject.keyword(), update_target_proj)
            with bff_struct(writer, is_keyword=True) as struct:
                struct.write_field('ProjectOutput', self.get_vcxproj_path(EDefaultTargets.Update.name))
                struct.write_field('OutputDirectory', PROJ_OUTPUT_DIR)

                struct.next()
                struct.next()
                write_assign(writer, 'GenerateProjectsCommands')
                with bff_struct(writer) as substruct:
                    substruct.write_field('ProjectBuildCommand', self.get_update_build_command(context))
                    substruct.write_field('ProjectRebuildCommand', self.get_update_rebuild_command(context))

                struct.next()
                struct.write_field('Configs', [])

                with bff_foreach(writer, '.ConfigMatrix'):
                    writer.newline()
                    write_assign(writer, 'ThisConfig')

                    with bff_struct(writer):
                        write_using(writer, 'Itr')
                        write_using(writer, 'GenerateProjectsCommands')

                    writer.newline()
                    write_append_field(writer, 'Configs', '.ThisConfig', outer_scope=True, no_quotations=True)

                writer.newline()
                struct.write_assign('ProjectConfigs', '.Configs')

            #
            # Default targets - commands
            #
            default_targets = [ all_target_proj, update_target_proj ]

            #
            # Solution
            #
            writer.newline()
            write_keyword(writer, sln.keyword(), f'{sln.Alias}')
            with bff_struct(writer, is_keyword=True) as struct:
                struct.write_field('SolutionOutput', f'{project_name}.sln')
                struct.write_field('SolutionProjects', default_targets + all_projects)
                struct.next()
                writer.write('.SolutionConfigs = .ConfigMatrix')
                struct.next()
                writer.write(f".SolutionBuildProject = '{all_target_proj}'")
                writer.newline()
                write_assign(writer, 'SolutionDependencies')
                with bff_struct(writer) as substruct:
                    substruct.write_field('Projects', all_projects)
                    substruct.write_field('Dependencies', [all_target_proj])

                min_ver, max_ver = get_min_max_msvc_version(context, context.project_template)
                struct.write_field('SolutionVisualStudioVersion', max_ver)
                struct.write_field('SolutionMinimumVisualStudioVersion', min_ver)

                # Folders
                folders = []
                def add_folder(name : str, title : str, targets : List[str]):
                    folders.append(f'.{name}')
                    writer.newline()
                    write_assign(writer, name)
                    with bff_struct(writer) as commands:
                        commands.write_field('Path', title)
                        commands.write_field('Projects', targets)

                writer.newline()
                add_folder('Folder_Commands', '0. Commands', default_targets)
                add_folder('Folder_Conan', '1. Dependencies', [ f'{target.target}-proj' for target in context.conan.dependencies ])
                add_folder('Folder_Libraries', '2. Libraries', [ f'{target.target}-proj' for target in filter(filter_target_libraries, context.project_template.targets) ])
                add_folder('Folder_Apps', '3. Apps', [ f'{target.target}-proj' for target in filter(filter_target_executable, context.project_template.targets) ])

                writer.newline()
                write_field(writer, 'SolutionFolders', folders, no_quotations=True)

            context.write_file(self.main_file_name, writer.to_string())

        self.emit_runner_script(context)

        return self.main_file_name


    def emit_target(self, context: EmitContext, target: ResolvedTarget):
        target_name = target.target.target
        target_file_name = f'{target_name}.bff'
        target_file_path = f'{FBUILD_OUTPUT_DIR}/{target_name}/{target_file_name}'
        project_configs_list = f'{target_name}_ProjectConfigs'

        project_paths = []
        config_includes = []

        if target.target.config.kind == ETargetKind.EXECUTABLE:
            config_includes = self.emit_executable(context, target, target_name, project_configs_list)

        elif target.target.config.kind == ETargetKind.DYNAMIC_LIBRARY:
            config_includes = self.emit_dynamic_library(context, target, target_name, project_configs_list)

        elif target.target.config.kind == ETargetKind.STATIC_LIBRARY:
            config_includes = self.emit_static_library(context, target, target_name, project_configs_list)

        elif target.target.config.kind == ETargetKind.IMPORTED_TARGET or target.target.config.kind == ETargetKind.HEADER_ONLY:
            config_includes = self.emit_imported_target(context, target, target_name, project_configs_list)

        elif target.target.config.kind == ETargetKind.BUILD_STEP:
            config_includes = self.emit_custom_build_step(context, target, target_name, project_configs_list)

        else:
            raise NotImplementedError(f"FASTBuild bff emission for target kind '{target.target.config.kind.name}' is not yet implemented!")

        vcxproj = vcxproject.VCXProject()
        vcxproj.Alias = f'{target_name}-proj'
        vcxproj.ProjectOutput = self.get_vcxproj_path(target_name)
        vcxproj.ProjectConfigs = project_configs_list
        vcxproj.ProjectBasePath = self.get_target_output_path(context, target)
        vcxproj.ProjectInputPaths = [vcxproj.ProjectBasePath] + project_paths
        vcxproj.ProjectFiles = [
            # f'{FBUILD_OUTPUT_DIR}/{target_name}/{target_file_name}',
            self.project_to_output_path(context, target.target.get_loaded_from_file())
        ]

        writer = create_bff_writer()
        self.emit_vcxproj(writer, vcxproj, project_configs_list, config_includes)

        context.write_file(target_file_path, writer.to_string())
        self.subprojects.append(target_file_path)



from argparse import ArgumentParser
import os
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.commands.execute_stage import execute_stage
from zetsubou.conan.dependencies import ConanDependencies
from zetsubou.project.model.toolchain_profile import Windows_ToolchainProfile
from zetsubou.project.model.virtual_environment import generate_empty_environment
from zetsubou.utils import logger
from zetsubou.utils.error_codes import EErrorCode
from zetsubou.utils.yaml_tools import to_yaml
from zetsubou.conan.conan import check_conan_present, call_conan
from zetsubou.conan.to_conan import runtime_to_conan, toolchain_to_conan, platform_to_conan, config_base_to_conan
from zetsubou.conan.deployer import get_conan_deployer_path


CONAN_OUT_PATH = 'build/conan'


class Install(Command):
    @property
    def name(self):
        return 'install'

    @property
    def desc(self):
        return 'Setup virtual environment, install build dependencies.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    def OnExecute(self, context: CommandContext):
        context.project_fs.makedirs('build/venv', recreate=True)

        if context.project_template.project.conan is not None and check_conan_present():
            self.generate_conan_environment(context)
        else:
            generate_empty_environment(context.project_fs)

    def merge_generated_part_files(self, context:CommandContext, conan_deps:str):
        dep_matrix = {}
        for path in context.project_fs.walk.files(path=CONAN_OUT_PATH, filter=['*.part.yml']):
            dep_name = os.path.basename(path).rsplit('.')[0]

            config_string = (os.path.basename(os.path.dirname(path))).replace('.', '-')
            if not context.config_matrix.has_variant(config_string):
                logger.Verbose(f"Conan dependency '{dep_name}' uses '*' fallback for all variants")
                config_string = '*'

            if dep_name not in dep_matrix:
                dep_matrix[dep_name] = [
                    f'target: {dep_name}\n'
                    'config:\n'
                    '  kind: IMPORTED_TARGET\n'
                    'filters:\n'
                ]

            dep_target = dep_matrix[dep_name]

            with context.project_fs.open(path, mode='r') as part_file:
                dep_target.append(f"  - filter:\n       config_string: '{config_string}'\n")
                dep_target.append(part_file.read())

            dep_matrix[dep_name] = dep_target

        # Merge targets configurations into one file
        out_ymls = []
        for dep_name, dep_content in dep_matrix.items():
            out_yml_path = f'{CONAN_OUT_PATH}/{dep_name}.yml'
            out_ymls.append(out_yml_path)

            with context.project_fs.open(out_yml_path, mode='w') as out_file:
                out_file.write(''.join(dep_content))

        # Merge targets into one file
        if len(out_ymls) > 0:

            with context.project_fs.open('build/conan_deps.yml', mode='w') as out_file:
                conan_deps = ConanDependencies(conanfile=conan_deps, targets=out_ymls)
                out_file.write(to_yaml(conan_deps))

            context.conan.yml_files = out_ymls

    def generate_conan_environment(self, context: CommandContext):
        conan_build_tools = context.project_template.project.conan.build_tools
        if conan_build_tools is not None:
            conan_build_tools = context.to_out_path(conan_build_tools)
            conan_out_directory = context.to_out_path("build/venv")
            execute_stage(lambda: call_conan(['install', conan_build_tools, '-g=VirtualBuildEnv', f'-of={conan_out_directory}'], conan_out_directory),
                        'Conan build tools installed',
                        EErrorCode.UNABLE_TO_INSTALL_CONAN_BUILD_TOOLS)

            context.resolve_venv()

        conan_deps = context.project_template.project.conan.dependencies
        if conan_deps is None:
            return

        conan_deps = context.to_out_path(conan_deps)

        for config_variant in context.config_matrix.variants:
            logger.Verbose(f"Installing configuration '{config_variant.config_string}'")

            plat_name = config_variant.get_slot('platform')
            platform = context.project_template.find_platform(plat_name)
            toolchain = context.project_template.find_toolchain(plat_name, config_variant.get_slot('toolchain'))
            config = context.project_template.find_config(config_variant.get_slot('configuration'))

            conan_settings = []
            conan_settings.extend(platform_to_conan(context, platform))
            conan_settings.extend(toolchain_to_conan(context, toolchain))
            conan_settings.extend(config_base_to_conan(context, config))

            if isinstance(toolchain.profile, Windows_ToolchainProfile):
                conan_settings.extend(runtime_to_conan(config, toolchain))

            conan_config_out_path = f'{CONAN_OUT_PATH}/{config_variant.config_string.replace("-", ".")}'
            context.project_fs.makedirs(conan_config_out_path, recreate=True)

            conan_call_args = [
                'install', conan_deps, f'--build={context.project_template.project.conan.build}',
                '--deploy', get_conan_deployer_path(),
                f'-of={conan_config_out_path}'
            ]

            # pylint: disable=cell-var-from-loop
            execute_stage(lambda: call_conan(conan_call_args + conan_settings,
                        context.to_out_path(conan_config_out_path),
                        context.fs_venv),
                        f"Conan configuration '{config_variant.config_string}' installed",
                        EErrorCode.UNABLE_TO_INSTALL_CONAN_DEPENDENCIES)

        self.merge_generated_part_files(context, conan_deps)



from enum import Enum


class ETargetKind(Enum):
    INVALID = 0x0
    EXECUTABLE = 0x1
    STATIC_LIBRARY = 0x2
    DYNAMIC_LIBRARY = 0x3
    OBJECT_LIBRARY = 0x4
    HEADER_ONLY = 0x5
    IMPORTED_TARGET = 0x6
    BUILD_STEP = 0x7


def is_target_kind_linkable(kind : ETargetKind) -> bool:
    if kind in [ ETargetKind.HEADER_ONLY, ETargetKind.IMPORTED_TARGET, ETargetKind.BUILD_STEP ]:
        return False
    return True


def is_target_library(kind : ETargetKind) -> bool:
    return kind != ETargetKind.EXECUTABLE


def is_target_prebuild_step(kind : ETargetKind) -> bool:
    return kind == ETargetKind.BUILD_STEP



import sys
import traceback
from enum import Enum
from colorama import Fore, Style, Back
from colorama import init as colorama_init
from zetsubou.utils.error_codes import EErrorCode
from bentoudev.dataclass.base import EErrorFormat


class ELogLevel(Enum):
    Verbose = 0
    Info = 1
    Warning = 2
    Error = 3
    Critical = 4
    Silent = 5


class scope_level:
    level: ELogLevel
    prev: ELogLevel

    def __init__(self, level: ELogLevel):
        self.level = level

    def __enter__(self):
        self.prev = LoggerStorage.log_level
        SetLogLevel(self.level)

    def __exit__(self, exc_type, exc_val, exc_tb):
        SetLogLevel(self.prev)


class LoggerStorage:
    log_level : ELogLevel = ELogLevel.Info
    ide : bool = False


def Initialize():
    colorama_init()


def SetLogLevel(log_level : ELogLevel):
    LoggerStorage.log_level = log_level


def SetIde(mode:bool):
    LoggerStorage.ide = mode


def IsVisible(level : ELogLevel):
    if level.value < LoggerStorage.log_level.value:
        return False
    return True


def IsIde():
    return LoggerStorage.ide


def GetErrorFormat() -> EErrorFormat:
    return EErrorFormat.MSVC if IsIde() else EErrorFormat.Pretty


def _prefix():
    if LoggerStorage.ide:
        return ''
    else:
        return '\r'


def ReturnCode(code: EErrorCode):
    if code == 0x0:
        Success('Returned code 0')
    elif IsVisible(ELogLevel.Critical):
        print(f'{_prefix()}{Fore.RED}{Back.WHITE} Returned error code {code.value} - {code.name} {Style.RESET_ALL}')


def CriticalError(err):
    if IsVisible(ELogLevel.Critical):
        print(_prefix())
        print(f'{_prefix()}{Fore.WHITE}{Back.RED} {err} {Style.RESET_ALL}')


def Exception(ex):
    if IsVisible(ELogLevel.Error):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        msg = 'Exception occured, '.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        Error(msg)


def Error(err):
    if IsVisible(ELogLevel.Error):
        print(_prefix())
        print(f'{_prefix()}{Fore.RED}{err}{Style.RESET_ALL}')


def Warning(warn):
    if IsVisible(ELogLevel.Warning):
        print(f'{_prefix()} - {Fore.YELLOW}{warn}{Style.RESET_ALL}')


def Info(inf):
    if IsVisible(ELogLevel.Info):
        print(f'{_prefix()} - {inf}')


def Success(suc):
    if IsVisible(ELogLevel.Info):
        print(f'{_prefix()} - [{Fore.GREEN}ok{Style.RESET_ALL}] {suc}')


def Loading(type: str, name: str):
    if IsVisible(ELogLevel.Info):
        print(f'{_prefix()} - Loading [{Fore.LIGHTBLUE_EX}{type}{Style.RESET_ALL}] \'{name}\'')


def Command(name: str):
    if IsVisible(ELogLevel.Info):
        print(f'{_prefix()}\n - Command [{Fore.CYAN}{name}{Style.RESET_ALL}]')


def Verbose(verb):
    if IsVisible(ELogLevel.Verbose):
        print(f'{_prefix()} - {verb}')



import argparse
from typing import List
from textwrap import dedent

from zetsubou._version import __version__, get_author_desc
from zetsubou.commands.base_command import Command


class SmartFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = dedent(text)
        return ''.join(indent + line for line in text.splitlines(True))


def parse_arguments(in_args:List[str], program_name: str, desc: str, command_registry : List[Command]):
    main_parser = argparse.ArgumentParser(prog=program_name, description=desc, formatter_class=SmartFormatter)

    main_parser.add_argument('--version', action='version', version=f'Zetsubou {__version__} - Copyright {get_author_desc()}')

    parent_parser = argparse.ArgumentParser(add_help=False)

    def add_global_options(parser):
        parser.add_argument('--nologo', action='store_true', default=False, help='skip printing of the logo')
        parser.add_argument('--ide', action='store_true', default=False, help='print messages in IDE friendly way')
        parser.add_argument('--dry', action='store_true', default=False, help='make no changes in filesystem')

        parser.add_argument('--profile', help='Constraints variant matrix to external profile')

        log_group = parser.add_mutually_exclusive_group()
        log_group.add_argument('--verbose', action='store_true', default=False, help='print more information during execution')
        log_group.add_argument('--silent', action='store_true', default=False, help='print only errors')

    add_global_options(parent_parser)

    subcommands = main_parser.add_subparsers(dest='command', metavar='COMMAND', required=True)

    for command in command_registry:
        sub_cmd = subcommands.add_parser(name=command.name, help=command.help, description=command.desc, parents=[parent_parser])
        if (command.needs_project_loaded):
            sub_cmd.add_argument('project', metavar='PROJECT', type=str, help='YAML file describing the project', default='project.yml')
        command.ParseArgs(sub_cmd)

    result = main_parser.parse_args(in_args)
    return result



from argparse import ArgumentParser
from typing import List, Optional
import os

import bentoudev.dataclass.base as base

from zetsubou.commands.execute_stage import execute_stage
from zetsubou.commands.command_context import CommandContext
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.runtime.project_loader import ProjectTemplate, load_project_from_file, ExternalConfig
from zetsubou.project.config_matrix import ConfigMatrix, ConfigVariant, create_config_matrix
from zetsubou.project.runtime.validation import ValidationContext, validate_profile
from zetsubou.system import discover_toolchains, tools
from zetsubou.utils import logger
from zetsubou.utils.common import fix_path
from zetsubou.utils.error_codes import EErrorCode


class Command:
    _commands : List['Command'] = []

    @staticmethod
    def is_initialized() -> bool:
        return len(Command._commands) > 0

    @staticmethod
    def initialize_command_instances(cmds : List['Command']):
        Command._commands = cmds

    @staticmethod
    def get_commands():
        return Command._commands

    @staticmethod
    def get_command_by_name(name:str):
        for c in Command._commands:
            if c.name == name:
                return c
        raise ValueError(f"Cannot find command with name '{name}'!")

    @staticmethod
    def get_command_instance(t):
        for c in Command._commands:
            if isinstance(c, t):
                return c
        raise ValueError(f"Cannot find instance of command '{t.__name__}'!")

    # User readable name
    @property
    def name(self):
        raise NotImplementedError()

    # Short information, displayed for --help
    @property
    def help(self):
        raise NotImplementedError()

    # Detailed information, displayed for COMMAND --help
    @property
    def desc(self):
        raise NotImplementedError()

    # Will require positional PROJECT argument
    # Will automatically load project before executing this command
    @property
    def needs_project_loaded(self):
        return True

    # Parses additional arguments that may be provided for this command
    def ParseArgs(self, arg_parser : ArgumentParser):
        raise NotImplementedError()

    def Execute(self, context : CommandContext):
        if self.needs_project_loaded:
            self._load_project(context)

        logger.Command(self.name)
        self.OnExecute(context)

    def OnExecute(self, context : CommandContext):
        raise NotImplementedError()

    def _load_project(self, context : CommandContext):
        if context.project_template is None:

            logger.Command('load_project')

            external_config = ExternalConfig()

            # handle platform from external file
            if hasattr(context.command_args, 'profile') and context.command_args.profile is not None:
                profile_file = os.path.relpath(context.command_args.profile, context.fs_root) if os.path.isabs(context.command_args.profile) else context.command_args.profile
                external_config.profile_file = fix_path(profile_file)

                logger.Info(f"Using profile from external file '{external_config.profile_file}'")

            # load and process project
            context.project_template = execute_stage(lambda: load_project_from_file(
                                        context.project_fs,
                                        context.fs_root,
                                        context.project_file,
                                        external_config=external_config),
                                        'Project loaded',
                                        EErrorCode.UNABLE_TO_LOAD_PROJECT)

            context.project_template.platforms = tools.filter_platforms_by_host(
                context.host_system,
                context.host_arch,
                context.project_template.platforms)

            # validate platforms
            vctx = ValidationContext()
            vctx.error_format = base.EErrorFormat.MSVC if context.command_args.ide else base.EErrorFormat.Pretty
            validate_profile(vctx, context.project_template.profile)

            # discovering and filtering of toolchains
            context.project_template.platform_toolchains, context.project_template.system_toolchains = execute_stage(
                lambda: discover_toolchains.discover_toolchain_list(context),
                'Toolchains discovered',
                EErrorCode.UNABLE_TO_DISCOVER_TOOLCHAIN)

            # create matrix of all options and their possible values
            context.config_matrix = execute_stage(lambda: resolve_config_matrix(context.project_template),
                                      'Config matrix created',
                                      EErrorCode.UNABLE_TO_CREATE_CONFIGURATIONS)


def resolve_config_matrix(project_template:ProjectTemplate) -> Optional[ConfigMatrix]:
    # Full config matrix, contains incorrect entries (toolchain/platform mismatched)
    config_matrix = create_config_matrix(project_template.project.config.config_string,
        project_template.slot_values())

    def filter_tool_by_plat(variant: ConfigVariant):
        toolchain = project_template.find_toolchain(
            variant.get_slot(EDefaultConfigSlots.platform.name),
            variant.get_slot(EDefaultConfigSlots.toolchain.name))

        return toolchain is not None

    # Filter out incorrect variants
    config_matrix.variants = list(filter(filter_tool_by_plat, config_matrix.variants))
    if len(config_matrix.variants) == 0:
        return None

    return config_matrix



from typing import Dict, List, Tuple, Optional

from zetsubou.commands.command_context import CommandContext
from zetsubou.conan.to_conan import toolsets_from_settings, msvc_version_to_conan
from zetsubou.conan.conan import get_conan_settings
from zetsubou.project.model.platform import Platform
from zetsubou.project.model.platform_enums import EArch, ESystem, EVersionSelector
from zetsubou.project.model.toolchain import Toolchain, ToolchainDefinition
from zetsubou.project.model.toolchain_enums import ECompilerFamily
from zetsubou.project.model.toolchain_profile import IToolchainProfile, Windows_ToolchainProfile
from zetsubou.utils import logger
from zetsubou.utils.common import is_in_enum


def is_second_newer_toolchain(first:ToolchainDefinition, second:ToolchainDefinition) -> bool:
    if first.CompilerFamily == ECompilerFamily.MSVC:
        if first.ide_version < second.ide_version:
            return True

    return first.version < second.version


# MSVC entries need to reparse due to how conan handles them
# Egx. 19.35 becomes 193.5 on conan
def is_matching_compiler_version(toolchain:ToolchainDefinition, versions:List[str]) -> bool:
    if toolchain.CompilerFamily == ECompilerFamily.MSVC:
        if msvc_version_to_conan(toolchain.version)[0] in versions:
            return True
    else:
        if str(toolchain.version.major) in versions:
            return True


class _ToolchainResult:
    definition: ToolchainDefinition
    filter: IToolchainProfile

    def __init__(self, defs:ToolchainDefinition, fltr:IToolchainProfile):
        self.definition = defs
        self.filter = fltr


def discover_toolchain_list(context: CommandContext) -> Optional[Tuple[Dict[str, List[Toolchain]], List[ToolchainDefinition]]] :
    system_toolchains   : List[ToolchainDefinition]  = discover_toolchain_by_system(context.host_system, context.host_arch)
    platform_toolchains : Dict[str, List[Toolchain]] = {}

    if len(system_toolchains) == 0:
        logger.CriticalError(f"No system toolchains detected on {context.host_system} {context.host_arch}!")
        return None

    conan_settings = get_conan_settings()

    # There were severe issues which toolchains on builders, so this code is left for convinience
    enable_toolchain_trace = False
    def trace_skipped_toolchain(tool_def : ToolchainDefinition, reason : str):
            logger.Verbose(f"TRACE: skipped toolchain '{tool_def.name}' ver '{tool_def.version}' reason: {reason}")

    if enable_toolchain_trace:
        logger.SetLogLevel(logger.ELogLevel.Verbose)

    for platform in context.project_template.platforms:
        compatible_toolchains = set()

        for tool_filter in context.project_template.profile.toolchains:
            # Single filter might output many toolchains (egx. with version set to all)
            latest_by_compiler : Optional[_ToolchainResult] = None

            logger.Verbose(f'Filter {tool_filter.compiler_family} {tool_filter.compiler_version} {tool_filter.target_arch} {tool_filter.toolset}')

            for tool_def in system_toolchains:
                if not is_in_enum(tool_def.arch, EArch):
                    trace_skipped_toolchain(tool_def, 'Unknown arch')
                    continue # Unknown arch

                if EArch[tool_def.arch] != platform.target_arch:
                    trace_skipped_toolchain(tool_def, 'Arch mismatch')
                    continue # Arch mismatch

                if tool_def.CompilerFamily != tool_filter.compiler_family:
                    trace_skipped_toolchain(tool_def, 'Family mismatch')
                    continue # Family mismatch

                if tool_filter.compiler_version == EVersionSelector.latest:
                    if latest_by_compiler is None:
                        latest_by_compiler = _ToolchainResult(tool_def, tool_filter)
                    elif is_second_newer_toolchain(latest_by_compiler.definition, tool_def):
                        latest_by_compiler = _ToolchainResult(tool_def, tool_filter)
                    trace_skipped_toolchain(tool_def, 'Not latest version')
                    continue

                elif not is_matching_compiler_version(tool_def, tool_filter.compiler_version):
                    trace_skipped_toolchain(tool_def, 'Version mismatch')
                    continue # Version mismatch

                compatible_toolchains.add(_ToolchainResult(tool_def, tool_filter))

            if latest_by_compiler is not None:
                compatible_toolchains.add(latest_by_compiler)

        def result_to_toolchain(tres:_ToolchainResult) -> List[Toolchain]:
            toolchains = []

            filter_by_toolset = isinstance(tres.filter, Windows_ToolchainProfile)
            if filter_by_toolset:

                def_toolsets = tres.definition.Toolset
                conan_toolsets = toolsets_from_settings(conan_settings, tres.definition.CompilerFamily)
                if conan_toolsets is not None:
                    def_toolsets = conan_toolsets.intersection(def_toolsets)

                if tres.filter.toolset == EVersionSelector.latest:
                    latest_toolset = max(def_toolsets)
                    tool = Toolchain(f"{tres.definition.name}_{latest_toolset}", tres.filter, tres.definition, toolset=latest_toolset)
                    toolchains.append(tool)

                else:
                    for toolset in tres.definition.Toolset:
                        tool = Toolchain(f"{tres.definition.name}_{toolset}", tres.filter, tres.definition, toolset=toolset)
                        toolchains.append(tool)

            else:
                toolchains.append(Toolchain(tres.definition.name, tres.filter, tres.definition, ''))

            return toolchains

        toolchain_objs = set()
        for res in map(result_to_toolchain, compatible_toolchains):
            toolchain_objs = toolchain_objs.union(res)

        platform_toolchains[platform.platform] = list(toolchain_objs)

    def system_toolchains_to_string():
        toolchains_labels = []
        for tool_def in system_toolchains:
            toolchains_labels.append(f"{tool_def.name} ver={tool_def.version}")

        prefix = '\n  - '
        label = prefix.join(toolchains_labels)
        return f"Available system toolchains:{prefix}{label}\n"

    if len(platform_toolchains) == 0:
        logger.CriticalError(f"No platform toolchains detected on {context.host_system} {context.host_arch}!")
        logger.Error(system_toolchains_to_string())
        return None

    for plat, tools in platform_toolchains.items():
        if len(tools) == 0:
            logger.CriticalError(f"No toolchains on platform '{plat}' detected on {context.host_system} {context.host_arch}!")
            logger.Error(system_toolchains_to_string())
            return None

    return (platform_toolchains, system_toolchains)


def discover_toolchain_by_system(host_system: ESystem, host_arch: EArch) -> List[ToolchainDefinition] :
    if host_system == ESystem.Windows:
        from zetsubou.system.windows import windows
        return windows.discover_toolchain_list(host_arch)
    else:
        raise EnvironmentError(f'Sorry, discovering toolchains on \"{host_system.name}\" arch \"{host_arch.name}\" is not currently supported!')


def filter_toolchains_by_target_arch(toolchain:ToolchainDefinition, platforms:List[Platform]):
    for plat in platforms:
        if not is_in_enum(toolchain.arch, EArch):
            logger.Warning(f"Unknown arch '{toolchain.arch}' for toolchain '{toolchain.name}'")
            return False

        if EArch[toolchain.arch] == plat.target_arch:
            return True

    return False



from argparse import ArgumentParser
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.fastbuild.fastbuild_emit import FastbuildEmitter
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.runtime.emit import EDefaultTargets, EmitContext
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.utils import logger
from zetsubou.utils.common import null_or_empty
from zetsubou.utils.error import CompilationError, ProjectError
from zetsubou.utils.subprocess import call_process_venv


class Build(Command):
    @property
    def name(self):
        return 'build'

    @property
    def desc(self):
        return 'Builds specified target for specified or all configs.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        config_or_variant = arg_parser.add_mutually_exclusive_group()

        config_or_variant.add_argument('--config', help='Builds all targets using this specific config')
        config_or_variant.add_argument('--target', help='Builds all configs for this specific target')
        config_or_variant.add_argument('--target-variant', help='Builds only this specific target variant')

    @property
    def needs_project_loaded(self):
        return True

    def pick_variants(self, context: CommandContext, arg_config:str):
        config_variants = []

        if not null_or_empty(arg_config):
            proj_config = context.project_template.find_config(arg_config)
            if proj_config is None:
                logger.Error(f"Unknown config '{arg_config}'")

            config_variants = context.config_matrix.find_variants({ 'configuration' : proj_config.configuration })

        else:
            config_variants = context.config_matrix.variants

        return config_variants

    def pick_target(self, context:CommandContext, arg_target:str) -> str:
        if null_or_empty(arg_target) or arg_target == EDefaultTargets.All.name:
            return EDefaultTargets.All.name

        build_target = context.project_template.find_target(arg_target)
        if build_target is None:
            raise ProjectError(f"Unknown target '{arg_target}'")

        return context.project_template.compile_target_variant_name(build_target.config.kind, build_target.target)

    def OnExecute(self, context: CommandContext):
        context.resolve_venv()

        try:
            arg_target = context.command_args.target
            arg_config = context.command_args.config
            arg_variant = context.command_args.target_variant

            target_variants = []

            if not null_or_empty(arg_variant):
                target_and_kind, config = ProjectTemplate.decompose_variant_name(arg_variant, context.project_template.project.config.config_string)
                if target_and_kind is None or config is None:
                    raise ProjectError(f"Unable to parse target variant string '{arg_variant}'")

                # TODO make this more robust
                target_parts = target_and_kind.split('_')
                if len(target_parts) > 2:
                    kind = f"{target_parts[1]}_{target_parts[2]}"
                    if not any(filter(lambda e : kind.startswith(e[0]), ETargetKind.__members__)):
                        raise ProjectError(f"Unknown target kind '{kind}'")

                target = target_parts[0]

                if not context.project_template.find_target(target) and target != EDefaultTargets.All.name:
                    raise ProjectError(f"Unknown target '{target}'")

                if not context.project_template.find_config(config[EDefaultConfigSlots.configuration.name]):
                    raise ProjectError(f"Unknown configuration '{config[EDefaultConfigSlots.configuration.name]}'")

                if not context.project_template.find_platform(config[EDefaultConfigSlots.platform.name]):
                    raise ProjectError(f"Unknown platform '{config[EDefaultConfigSlots.platform.name]}'")

                if not context.project_template.find_toolchain(config[EDefaultConfigSlots.platform.name], config[EDefaultConfigSlots.toolchain.name]):
                    raise ProjectError(f"Unknown toolchain '{config[EDefaultConfigSlots.toolchain.name]}'")

                target_variants = [ arg_variant ]

            else:
                # We have four options
                # - target with config
                # - target with all configs
                # - all targets with config
                # - all targets with all configs

                config_variants = self.pick_variants(context, arg_config)
                target = self.pick_target(context, arg_target)

                # Problem with double/naming of matrices, since per config target files contain config name in them
                if target == EDefaultTargets.All.name:
                    target_variants = list(map(
                        lambda conf_var: f'{target}-{conf_var.config_string}',
                        config_variants
                    ))
                else:
                    target_variants = list(map(
                        lambda conf_var: context.project_template.compile_target_variant_name(ETargetKind.INVALID, target, conf_var.config_string),
                        config_variants
                    ))

            # Refactor context, ensure this makes sense
            emit_context = EmitContext(
                project_template=context.project_template,
                resolved_targets=context.resolved_targets,
                config_matrix=context.config_matrix,
                conan=context.conan,
                mem_fs=None,
                project_fs=None)

            # From fbuild emitter, funcs to find paths and build command strings
            for tar_var in target_variants:
                command_string = FastbuildEmitter.get_fbuild_build_command(emit_context, tar_var)
                logger.Info(f"Building '{tar_var}'...")

                args = command_string.split(' ')
                _, err = call_process_venv(args, context.fs_venv, capture = False, realtime = True, cwd = context.fs_root)
                if err is not None:
                    err_msg = f"FBuild returned error '{err}'"
                    logger.CriticalError(err_msg)
                    raise CompilationError(err_msg)

        except ProjectError:
            raise

        except CompilationError:
            raise

        except Exception as exc:
            logger.Exception(exc)
            raise



from typing import List
from dataclasses import dataclass, field

from zetsubou.project.model.configuration_enums import EBaseConfiguration
from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.target import TargetData
from bentoudev.dataclass.base import loaded_from_file


@dataclass
class ConfigurationData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
@loaded_from_file
class Configuration(TargetData):
    configuration: str = None
    base_configuration: EBaseConfiguration = None
    filters: List[ConfigurationData] = field(default_factory=list)



from typing import List
from dataclasses import dataclass
from bentoudev.dataclass.base import loaded_from_file, track_source

@dataclass
@track_source
@loaded_from_file
class ConanDependencies():
    conanfile: str
    targets: List[str]



from dataclasses import dataclass, fields
from typing import List, Optional
from abc import ABC, abstractmethod

from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.toolchain_enums import ECompilerFamily, ELibrarianFamily, ELinkerFamily, ECppStandard
from zetsubou.project.model.toolchain_profile import IToolchainProfile
from zetsubou.utils.common import Version, filename_no_ext, fix_path


class ICompiler(ABC):
    @abstractmethod
    def output(self) -> str:
        pass

    @staticmethod
    @abstractmethod
    def cpp_to_str(cpp:ECppStandard) -> Optional[str]:
        pass

    @abstractmethod
    def cppstd(self, cpp:ECppStandard) -> str:
        pass

    @abstractmethod
    def include(self, value:str) -> str:
        pass

    @abstractmethod
    def system_include(self, value:str) -> str:
        pass

    @abstractmethod
    def define(self, value:str) -> str:
        pass


class ClangCompiler(ICompiler):
    def output(self) -> str:
        return '-o"%2" "%1" -c '

    @staticmethod
    def cpp_to_str(cpp:ECppStandard) -> Optional[str]:
        return {
            ECppStandard.cpp11  : 'c++11',
            ECppStandard.cpp14  : 'c++14',
            ECppStandard.cpp17  : 'c++17',
            ECppStandard.cpp20  : 'c++20',
            ECppStandard.cpp23  : 'c++2b',
            ECppStandard.latest : 'c++2b',
        }.get(cpp, None)

    def cppstd(self, cpp: ECppStandard) -> str:
        return f"-std={ClangCLCompiler.cpp_to_str(cpp)}"

    def include(self, value:str) -> str:
        return f"-I\"{fix_path(value)}\""

    def system_include(self, value: str) -> str:
        return f"-isystem\"{fix_path(value)}\""

    def define(self, value:str) -> str:
        return f"-D{value}"


class MSVCCompiler(ICompiler):
    def output(self) -> str:
        return '/c "%1" /Fo"%2" /experimental:external '

    @staticmethod
    def cpp_to_str(cpp:ECppStandard) -> Optional[str]:
        return {
            ECppStandard.cpp14  : 'c++14',
            ECppStandard.cpp17  : 'c++17',
            ECppStandard.cpp20  : 'c++20',
            ECppStandard.latest : 'latest',
        }.get(cpp, None)

    def cppstd(self, cpp: ECppStandard) -> str:
        return f"/std:{MSVCCompiler.cpp_to_str(cpp)} "

    def include(self, value:str) -> str:
        return f"/I\"{value}\""

    def system_include(self, value:str) -> str:
        return f"/external:I\"{value}\""

    def define(self, value:str) -> str:
        return f"/D{value}"


class ClangCLCompiler(MSVCCompiler):
    pass


class GCCCompiler(ClangCompiler):
    pass


def _pick_by_filename(entries, executable):
    filename = filename_no_ext(executable)
    for entry in entries:
        if filename in entry[1]:
            return entry[0]
    return None


def guess_compiler_family(compiler:str) -> Optional[ECompilerFamily]:
    return _pick_by_filename([
        (ECompilerFamily.MSVC,     ['cl']),
        (ECompilerFamily.CLANG,    ['clang', 'clang++']),
        (ECompilerFamily.CLANG_CL, ['clang-cl']),
        (ECompilerFamily.GCC,      ['gcc', 'g++']),
    ], compiler)


def get_compiler_by_family(family:ECompilerFamily) -> ICompiler:
    return {
        ECompilerFamily.MSVC : MSVCCompiler(),
        ECompilerFamily.CLANG : ClangCompiler(),
        ECompilerFamily.CLANG_CL : ClangCLCompiler(),
        ECompilerFamily.GCC : GCCCompiler(),
    }.get(family)


class ILibrarian(ABC):
    @abstractmethod
    def output(self) -> str:
        pass


class MSVC_Librarian(ILibrarian):
    def output(self) -> str:
        return '/OUT:"%2" "%1" '


class LLVM_Librarian(ILibrarian):
    def output(self) -> str:
        return 'rc "%2" "%1" '


class GCC_Librarian(LLVM_Librarian):
    pass


def guess_librarian_family(librarian:str) -> Optional[ELibrarianFamily]:
    return _pick_by_filename([
        (ELibrarianFamily.MSVC, ['lib']),
        (ELibrarianFamily.LLVM, ['llvm-ar']),
        (ELibrarianFamily.GCC,  ['ar']),
    ], librarian)


def get_librarian_by_family(family:ELibrarianFamily) -> ILibrarian:
    return {
        ELibrarianFamily.MSVC : MSVC_Librarian(),
        ELibrarianFamily.LLVM : LLVM_Librarian(),
        ELibrarianFamily.GCC : GCC_Librarian(),
    }.get(family)


class ILinker(ABC):
    @abstractmethod
    def output(self) -> str:
        pass

    @abstractmethod
    def target(self, target_kind: ETargetKind) -> str:
        pass

    @abstractmethod
    def link(self, value:str) -> str:
        pass

    @abstractmethod
    def dir(self, value:str) -> str:
        pass


class LINK_Linker(ILinker):
    def output(self) -> str:
        return '/OUT:"%2" "%1" '

    def target(self, target_kind: ETargetKind) -> str:
        if target_kind == ETargetKind.DYNAMIC_LIBRARY:
            return '/DLL '
        return ''

    def link(self, value: str) -> str:
        return value

    def dir(self, value: str) -> str:
        return f"/LIBPATH:\"{value}\""


class LD_Linker(ILinker):
    def output(self) -> str:
        return '"%1" -o "%2" '

    def target(self, target_kind: ETargetKind) -> str:
        if target_kind == ETargetKind.DYNAMIC_LIBRARY:
            return '--shared '
        return ''

    def link(self, value: str) -> str:
        return f"-l\"{value}\""

    def dir(self, value: str) -> str:
        return f"-L\"{fix_path(value)}\""


class LLD_LINK_Linker(LINK_Linker):
    pass


class LD_LLD_Linker(LD_Linker):
    pass


def guess_linker_family(linker:str) -> Optional[ELinkerFamily]:
    return _pick_by_filename([
        (ELinkerFamily.LINK,     ['link']),
        (ELinkerFamily.LLD_LINK, ['lld-link']),
        (ELinkerFamily.LD_LLD,   ['ld-lld']),
        (ELinkerFamily.LD,       ['ld']),
    ], linker)


def get_linker_by_family(family: ELinkerFamily) -> ILinker:
    return {
        ELinkerFamily.LINK : LINK_Linker(),
        ELinkerFamily.LLD_LINK : LLD_LINK_Linker(),
        ELinkerFamily.LD_LLD : LD_LLD_Linker(),
        ELinkerFamily.LD : LD_Linker(),
    }.get(family)


@dataclass
class ToolchainDefinition:
    name: str
    version : Version
    ide_version : Version
    vs_install_path : str
    arch : str

    i_compiler : ICompiler
    i_librarian : ILibrarian
    i_linker : ILinker

    # Toolset
    Toolset : List[str]

    # Directories added to PATH
    PathEnv: List[str]

    # Default include directories
    IncludeDirectories: List[str]

    # Additonal files copied with exe during distributed compilation
    RequiredFiles: List[str]

    # (optional) Additional files (usually dlls) required by the compiler.
    ExtraFiles: List[str]

    # Default defines
    Defines: List[str]



    # Compiler executable
    Compiler: str

    # Compiler flags
    CompilerOptions : str

    # Compiler family
    CompilerFamily: ECompilerFamily



    # Linker executable
    Linker: str

    # Linker family
    LinkerFamily : ELinkerFamily

    # Linker flags
    LinkerOptions : str

    # Paths with library files
    LinkerPaths : str



    # Librarian executable
    Librarian: str

    # Librarian family
    LibrarianFamily : ELibrarianFamily

    # Librarian flags
    LibrarianOptions : str



    @staticmethod
    def create(name:str, compiler_exe:str, librarian_exe:str, linker_exe:str,
     *, CompilerOptions: str = '', LibrarianOptions: str = '', LinkerOptions: str = '', **kwargs) -> Optional['ToolchainDefinition']:

        compiler_family = guess_compiler_family(compiler_exe)
        librarian_family = guess_librarian_family(librarian_exe)
        linker_family = guess_linker_family(linker_exe)

        i_compiler = get_compiler_by_family(compiler_family)
        i_librarian = get_librarian_by_family(librarian_family)
        i_linker = get_linker_by_family(linker_family)

        return ToolchainDefinition(
            name=name,

            i_compiler=i_compiler,
            Compiler=compiler_exe,
            CompilerFamily=compiler_family,
            CompilerOptions=i_compiler.output() + CompilerOptions,

            i_librarian=i_librarian,
            Librarian=librarian_exe,
            LibrarianFamily=librarian_family,
            LibrarianOptions=i_librarian.output() + LibrarianOptions,

            i_linker=i_linker,
            Linker=linker_exe,
            LinkerFamily=linker_family,
            LinkerOptions=i_linker.output() + LinkerOptions,

            **{ name : value for name, value in kwargs.items() }
        )

    def __hash__(self):
        return hash(self.name)

    def get_hidden_fields(self) -> List[str]:
        result = []
        for field in fields(ToolchainDefinition):
            if field.name.islower():
                result.append(field.name)
        return result


# The actual toolchain that can be used
@dataclass
class Toolchain:
    name: str
    profile: IToolchainProfile
    definition: ToolchainDefinition
    toolset: str

    def __hash__(self):
        return hash(self.name)



from typing import List

import bentoudev.dataclass.yaml_loader as yaml
import bentoudev.dataclass.base as base

from zetsubou.utils import logger
from zetsubou.utils.dataclass_loader import DataclassLoader

class YamlDataclassLoader(DataclassLoader):
    def __init__(self):
        self.type_cache = yaml.default_type_loaders()

    def load_dataclass(self, clazz: type, filename: str, file_content: str, ext_types: List[type], *, error_format=base.EErrorFormat.Pretty):
        classname = clazz.__name__.lower()
        label = f"[{classname}] {filename}"
        logger.Loading(classname, filename)

        try:
            return yaml.load_yaml_dataclass(clazz, label, file_content, type_cache=self.type_cache, ext_types=ext_types, always_track_source=True, error_format=error_format)

        except base.DataclassLoadError as error:
            logger.Error(error)
            logger.CriticalError(f'Failed to load {classname} \'{filename}\'')
            return None



from argparse import ArgumentParser
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.utils import logger
from zetsubou.utils.error_codes import EErrorCode, ReturnErrorcode


class Clean(Command):
    @property
    def name(self):
        return 'clean'

    @property
    def desc(self):
        return 'Removes build folder and solution files.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    @property
    def needs_project_loaded(self):
        return False

    def OnExecute(self, context: CommandContext):
        try:
            logger.Info("Removing 'build' ...")
            if context.project_fs.exists('build'):
                context.project_fs.removetree('build')

        except Exception as err:
            logger.Error(f"Failed to remove 'build'")
            raise ReturnErrorcode(EErrorCode.UNABLE_TO_CLEAN) from err

        # --

        try:
            logger.Info("Removing '*.sln' ...")
            for path in context.project_fs.walk(path='.', filter=['*.sln']):
                for file in path.files:
                    context.project_fs.remove(file.name)

        except Exception as err:
            logger.Error("Failed to remove '*.sln'")
            raise ReturnErrorcode(EErrorCode.UNABLE_TO_CLEAN) from err



from enum import Enum, auto


# class ERuntimeLibrary(Enum):
#     DYNAMIC_DEBUG   = 0,
#     DYNAMIC_RELEASE = 1,
#     STATIC_DEBUG    = 2,
#     STATIC_RELEASE  = 3


class ERuntimeLibrary(Enum):
    dynamic = auto()
    static = auto()



from typing import List, Optional
from dataclasses import dataclass, field


@dataclass
class Compiler:
    Executable : str
    ExtraFiles : List[str] = field(default_factory=list)

    CompilerFamily : str = 'auto'
    AllowDistribution : bool = True
    ExecutableRootPath : Optional[str] = None
    SimpleDistributionMode : Optional[bool] = None
    CustomEnvironmentVariables : Optional[List[str]] = field(default_factory=list)
    ClangRewriteIncludes : Optional[bool] = None
    ClangGCCUpdateXLanguageArg : Optional[bool] = None
    VS2012EnumBugFix : Optional[bool] = None
    Environment : Optional[List[str]] = field(default_factory=list)
    AllowResponseFile : Optional[bool] = None
    ForceResponseFile : Optional[bool] = None

    UseLightCache_Experimental : Optional[bool] = None
    UseRelativePaths_Experimental : Optional[bool] = None
    SourceMapping_Experimental : Optional[bool] = None
    ClangFixupUnity_Disable : Optional[bool] = None

    @staticmethod
    def keyword() -> str:
        return "Compiler"



from typing import Optional, List
from dataclasses import dataclass

from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.runtime_library import ERuntimeLibrary
from zetsubou.project.model.configuration_enums import EBaseConfiguration


@dataclass
class TargetFilter():
    config_string: str = '*'
    kind: Optional[List[ETargetKind]] = None
    runtime_library: Optional[List[ERuntimeLibrary]] = None
    base_configuration: Optional[List[EBaseConfiguration]] = None



from sys import exit as sys_exit
from sys import argv as sys_argv
from typing import Callable, List, Optional
from .zetsubou import main as zet_main


# class ProfileScope:
#     scalene = None

#     def __init__(self, argv:List[str]):
#         profile_arg = '--profile'
#         if profile_arg in argv:
#             argv.remove(profile_arg)
#             from scalene import scalene_profiler
#             self.scalene = scalene_profiler.Scalene
            

#     def __enter__(self):
#         if self.scalene is not None:
#             #self.scalene.clear_mmap_data()
#             self.scalene.__stats.start_clock()
#             self.scalene.enable_signals()

#     def __exit__(self, exc_type, exc_val, exc_tb):
#         if self.scalene is not None:
#             self.scalene.disable_signals()
#             self.scalene.__stats.stop_clock()


# with ProfileScope(sys_argv[1:]):

sys_exit(zet_main(sys_argv[1:]))



from typing import List, Optional
from zetsubou.project.model.cli_tool import CommandlineTool
from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.target import Dependencies, Target, TargetReference, find_target
from zetsubou.project.model.toolchain_profile import IToolchainProfile, Profile
from zetsubou.project.model.toolchain import ECppStandard, ICompiler, get_compiler_by_family
from zetsubou.utils.common import foreach
from zetsubou.utils.error import ProjectError

import bentoudev.dataclass.base as base


class ValidationContext:
    error_format: base.EErrorFormat


def find_unresolved_target_references(ctx:ValidationContext, targets: List[Target], deps: Optional[List[TargetReference]]) -> List[TargetReference]:
    failed_refs = []
    if deps:
        for ref in deps:
            found_target = find_target(targets, ref.name)
            if found_target is None:
                failed_refs.append(ref)
            else:
                ref.target = found_target
    return failed_refs


def validate_duplicates(ctx:ValidationContext, targets: List[Target]):
    visited = dict()
    for target in targets:
        if target.target in visited:
            tmp_msg = target.format_field_message('target', 'error', f"Duplicated target name '{target.target}'!", ctx.error_format)
            err_msg = f"{tmp_msg}\n'{visited[target.target].format_field_message('target', '', 'Second', ctx.error_format)}'"
            raise ProjectError(err_msg)
        else:
            visited[target.target] = target


def validate_cyclic_dependency(ctx:ValidationContext, targets: List[Target]):
    visited = dict()

    def visit_target(in_target):
        nonlocal visited

        target = in_target

        if isinstance(in_target, TargetReference):
            target = in_target.target

        if target.target in visited:
            target_path = []
            msg = []

            for path_element, path_ref in visited.items():
                target_path.append(path_element)
                target_path.append(' -> ')

                if isinstance(path_ref, TargetReference):
                    msg.append(path_ref.format_message('', '\nReferenced here', ctx.error_format))
                    msg.append('\n')

            target_path.append(f'{target.target}')

            path_msg = ''.join(target_path)
            ref_msg = ''.join(msg)

            main_msg = in_target.format_message('error', f"Cyclic dependency to target '{target.target}' through '{path_msg}'", ctx.error_format)
            err_msg = f'{main_msg}\n{ref_msg}'
            raise ProjectError(err_msg)

        def visit_dependencies(t: Target, d: Dependencies):
            foreach(d.interface, visit_target)
            foreach(d.public, visit_target)
            foreach(d.private, visit_target)

        visited[target.target] = in_target

        if target.dependencies is not None:
            visit_dependencies(target, target.dependencies)

        if target.filters is not None:
            for f in target.filters:
                if f.dependencies is not None:
                    visit_dependencies(target, f.dependencies)

        del visited[target.target]

    for target in targets:
        visit_target(target)


def validate_unresolved_target_dependencies(ctx:ValidationContext, target_name: str, dependencies, targets: List[Target]):
    if dependencies is not None:
        for name, dep_list in {
            'interface': dependencies.interface,
            'public': dependencies.public,
            'private': dependencies.private
        }.items():

            unresolved_refs = find_unresolved_target_references(ctx, targets, dep_list)

            if len(unresolved_refs) > 0:
                msg = ''
                for un_ref in unresolved_refs:
                    msg += un_ref.format_message('error', f"\nUnknown target '{un_ref.name}'", ctx.error_format)

                raise ProjectError(
                    f'Reference to undefined targets in \'{name}\' dependencies of target \'{target_name}\'{msg}')


def validate_targets_not_empty(ctx:ValidationContext, targets: List[Target]):
    if (len(targets)) == 0:
        raise ProjectError("Cannot load project with no targets!")


def validate_unresolved_targets(ctx:ValidationContext, targets: List[Target]):
    for target in targets:
        validate_unresolved_target_dependencies(ctx, target.target, target.dependencies, targets)
        for f in target.filters:
            validate_unresolved_target_dependencies(ctx, target.target, f.dependencies, targets)


def validate_custom_build_steps(ctx:ValidationContext, targets: List[Target]):
    for target in targets:
        if target.config.kind != ETargetKind.BUILD_STEP and target.config.compiler is not None:
            raise ProjectError(f'Custom compiler cannot be defined for target \'{target.target}\' of kind \'{target.config.kind.name}\'')
        elif target.config.kind == ETargetKind.BUILD_STEP and target.config.compiler is None:
            raise ProjectError(f'Custom compiler must be defined for target \'{target.target}\' of kind \'{target.config.kind.name}\'')


def validate_cli_tools(ctx:ValidationContext, targets : List[Target], cli_tools : List[CommandlineTool]):
    for target in targets:
        if target.config.kind == ETargetKind.BUILD_STEP:
            found_tool:bool = False
            for tool in cli_tools:
                if tool.name == target.config.compiler:
                    found_tool = True
                    break
            if not found_tool:
                raise ProjectError(f'Reference to undefined custom compiler \'{target.config.compiler}\' for target \'{target.target}\'')


def validate_targets(ctx:ValidationContext, targets : List[Target]):
    validate_targets_not_empty(ctx, targets)
    validate_unresolved_targets(ctx, targets)
    validate_duplicates(ctx, targets)
    validate_cyclic_dependency(ctx, targets)
    validate_custom_build_steps(ctx, targets)


##########################################################################


# def validate_compiler_duplicates(ctx:ValidationContext, platforms : List[Platform]):
#     for plat in platforms:
#         compiler_families = dict()
#         for tool_filter in plat.toolchains:
#             if tool_filter.family in compiler_families:
#                 first_occur = compiler_families[tool_filter.family].format_field_message('family', '', 'First referenced here', ctx.error_format)
#                 msg = tool_filter.format_field_message(
#                     'family',
#                     'error',
#                     f"Duplicated toolchain definition for '{tool_filter.family.name}'.\n{first_occur}", ctx.error_format)
#                 raise ProjectError(msg)
#             else:
#                 compiler_families[tool_filter.family] = tool_filter


def validate_cpp_standard(ctx:ValidationContext, toolchains : List[IToolchainProfile]):
    def built_supported_values(compiler_t:ICompiler):
        all_values = ECppStandard.__members__.items()
        allowed_values = []
        for value in all_values:
            if compiler_t.cpp_to_str(value[1]) is not None:
                allowed_values.append(value[0])
        return allowed_values

    for tool_filter in toolchains:
        compiler_t = get_compiler_by_family(tool_filter.compiler_family)
        if compiler_t.cpp_to_str(tool_filter.cppstd) is None:
            msg = tool_filter.format_field_message(
                    'cppstd',
                    'error',
                    f"Unsupported C++ version '{tool_filter.cppstd.name}' for compiler family '{tool_filter.compiler_family.name}'.\n"
                    f"Supported are: {', '.join(built_supported_values(compiler_t))}",
                    ctx.error_format)
            raise ProjectError(msg)


def validate_profile(ctx:ValidationContext, profile : Profile):
    # validate_compiler_duplicates(ctx, platforms)
    validate_cpp_standard(ctx, profile.toolchains)



from typing import List
from zetsubou.commands.base_command import Command
from zetsubou.commands.clean import Clean
from zetsubou.commands.generate import Generate
from zetsubou.commands.regen import Regen
from zetsubou.commands.install import Install
from zetsubou.commands.create import Create
from zetsubou.commands.build import Build
from zetsubou.commands.clean_build import CleanBuild
from zetsubou.commands.config import Config


def get_all_commands() -> List[Command]:
    if not Command.is_initialized():
        Command.initialize_command_instances([
            Clean(),
            Config(),
            Install(),
            Generate(),
            Regen(),
            Create(),
            Build(),
            CleanBuild(),
        ])

    return Command.get_commands()



from argparse import ArgumentParser
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.commands.execute_stage import execute_stage
from zetsubou.utils import logger

import fs, os
from fs.copy import copy_fs
from fs.base import FS
from fs.memoryfs import MemoryFS
from fs.osfs import OSFS
from zetsubou.utils.error_codes import EErrorCode


def emit_project_template(templ_dir:str, data_fs:FS, out_fs:FS):
    if not data_fs.exists(templ_dir) or not data_fs.isdir(templ_dir):
        return None

    for itr in data_fs.walk(templ_dir):
        rel_path = fs.path.relativefrom(templ_dir, itr.path)
        out_fs.makedirs(rel_path, recreate=True)
        for file in itr.files:
            rel_file = fs.path.join(rel_path, file.name)
            out_file_handle = out_fs.openbin(rel_file, mode='w')
            data_fs.download(fs.path.join(itr.path, file.name), out_file_handle)

    return out_fs


class Create(Command):
    @property
    def name(self):
        return 'create'

    @property
    def desc(self):
        return 'Creates new project from selected template.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    @property
    def needs_project_loaded(self):
        return False

    def OnExecute(self, context: CommandContext):
        try:
            project_template = 'data/default_project'

            out_fs = execute_stage(lambda: emit_project_template(project_template, data_fs=OSFS(os.getcwd()), out_fs=MemoryFS()),
                'Created new project',
                EErrorCode.UNABLE_TO_COMMIT_FS)

            if out_fs is not None and not context.command_args.dry:
                copy_fs(out_fs, context.project_fs)

        except Exception as exc:
            logger.Exception(exc)
            raise



from typing import List, Optional


class BaseTarget:
    Alias : str

    @staticmethod
    def keyword() -> str:
        return NotImplementedError()


class LinkableTarget(BaseTarget):

    # Linker executable to use
    Linker : str
    # Output from linker
    LinkerOutput : str
    # Options to pass to linker
    LinkerOptions : str
    # Libraries to link into DLL
    Libraries : List[str]
    # Secondary libraries to link into executable
    Libraries2 : List[str]
    # (optional) Link objects used to make libs instead of libs (default true)
    LinkerLinkObjects : Optional[str]
    # (optional) List of assembly resources to use with %3
    LinkerAssemblyResources : Optional[str]

    # (optional) Executable to run post-link to "stamp" executable in-place
    LinkerStampExe : Optional[str]
    # (optional) Arguments to pass to LinkerStampExe
    LinkerStampExeArgs : Optional[str]
    # (optional) Specify the linker type. Valid options include: 
    # auto, msvc, gcc, snc-ps3, clang-orbis, greenhills-exlr, codewarrior-ld
    # Default is 'auto' (use the linker executable name to detect)
    LinkerType : Optional[str]

    # (optional) Allow response files to be used if not auto-detected (default: false)
    LinkerAllowResponseFile : Optional[bool]
    # (optional) Force use of response files (default: false)
    LinkerForceResponseFile : Optional[bool]

    ###############################################
    # Additional options
    ###############################################

    # (optional) Force targets to be built before this DLL (Rarely needed,
    # but useful when DLL relies on externally generated files).
    PreBuildDependencies : List[str]
    # (optional) Environment variables to use for local build
    # If set, linker uses this environment
    # If not set, linker uses .Environment from your Settings node
    Environment: List[str]


class AppTarget(LinkableTarget):
    @staticmethod
    def keyword() -> str:
        return "Executable"


class DynLibTarget(LinkableTarget):
    @staticmethod
    def keyword() -> str:
        return "DLL"


class ObjTarget(BaseTarget):
    # (optional) Alias
    Alias : Optional[str]

    ###############################################
    # Options for compilation
    ###############################################

    # Compiler to use
    Compiler : str
    # Options for compiler
    CompilerOptions : str
    # Path to store intermediate objects
    CompilerOutputPath : str
    # (optional) Specify the file extension for generated objects (default .obj or .o)
    CompilerOutputExtension : Optional[str]
    # (optional) Append extension instead of replacing it (default: false)
    CompilerOutputKeepBaseExtension : Optional[bool]
    # (optional) Specify a prefix for generated objects (default none)
    CompilerOutputPrefix : Optional[str]

    ###############################################
    # Specify inputs for compilation
    ###############################################

    # (optional) Path to find files in
    CompilerInputPath : Optional[str]
    # (optional) Pattern(s) to use when finding files (default *.cpp)
    CompilerInputPattern : Optional[str]
    # (optional) Recurse into dirs when finding files (default true)
    CompilerInputPathRecurse : Optional[bool]
    # (optional) Path(s) to exclude from compilation
    CompilerInputExcludePath : Optional[str]
    # (optional) File(s) to exclude from compilation (partial, root-relative of full path)
    CompilerInputExcludedFiles : Optional[str]
    # (optional) Pattern(s) to exclude from compilation
    CompilerInputExcludePattern : Optional[str]
    # (optional) Explicit array of files to build
    CompilerInputFiles : Optional[List[str]]
    # (optional) Root path to use for .obj path generation for explicitly listed files
    CompilerInputFilesRoot : Optional[str]
    # (optional) Unity to build (or Unities)
    CompilerInputUnity : Optional[str]
    # (optional) Don't fail if no inputs are found
    CompilerInputAllowNoFiles : Optional[bool]
    # (optional) ObjectList(s) whos output should be used as an input
    CompilerInputObjectLists : Optional[str]

    # (optional) Allow caching of compiled objects if available (default true)
    AllowCaching : Optional[bool]
    # (optional) Allow distributed compilation if available (default true)
    AllowDistribution : Optional[bool]

    # (optional) Compiler to use for preprocessing
    Preprocessor : Optional[str]
    # (optional) Args to pass to compiler if using custom preprocessor
    PreprocessorOptions : Optional[str]

    # (optional) List of objects to be used with /FU
    CompilerForceUsing : Optional[str]

    ###############################################
    # (optional) Properties to control precompiled header use
    ###############################################

    # (optional) Precompiled header (.cpp) file to compile
    PCHInputFile : Optional[str]
    # (optional) Precompiled header compilation output
    PCHOutputFile : Optional[str]
    # (optional) Options for compiler for precompiled header
    PCHOptions : Optional[str]

    ###############################################
    # Additional options
    ###############################################

    # (optional) Force targets to be built before this library (Rarely needed,
    # but useful when a library relies on generated code).
    PreBuildDependencies : Optional[str]

    # (optional) Hide a target from -showtargets (default false)
    Hidden : Optional[bool]

    @staticmethod
    def keyword() -> str:
        return "ObjectList"


class StaticLibTarget(BaseTarget):

    ###############################################
    # Options for compilation
    ###############################################
    # Compiler to use
    Compiler : str
    # Options for compiler
    CompilerOptions : str
    # Path to store intermediate objects
    CompilerOutputPath : str
    # (optional) Specify the file extension for generated objects (default .obj or .o)
    CompilerOutputExtension : Optional[str]
    # (optional) Specify a prefix for generated objects (default none)
    CompilerOutputPrefix : Optional[str]

    ###############################################
    # Options for librarian
    ###############################################

    # Librarian to collect intermediate objects
    Librarian : str

    # Options for librarian
    LibrarianOptions : str

    # (optional) Specify the librarian type. Valid options include:
    # auto, msvc, ar, ar-orbis, greenhills-ax
    # Default is 'auto' (use the librarian executable name to detect)
    LibrarianType : Optional[str]

    # Output path for lib file
    LibrarianOutput : str

    # (optional) Additional inputs to merge into library
    LibrarianAdditionalInputs : Optional[str]

    # (optional) Allow response files to be used if not auto-detected (default: false)  
    LibrarianAllowResponseFile : Optional[bool]

    # (optional) Force use of response files (default: false)
    LibrarianForceResponseFile : Optional[bool]

    ###############################################
    # Specify inputs for compilation
    ###############################################
    # (optional) Path to find files in
    CompilerInputPath : Optional[str]

    # (optional) Pattern(s) to use when finding files (default *.cpp)
    CompilerInputPattern : Optional[List[str]]
    # (optional) Recurse into dirs when finding files (default true)
    CompilerInputPathRecurse : Optional[bool]
    # (optional) Path(s) to exclude from compilation
    CompilerInputExcludePath : Optional[List[str]]
    # (optional) File(s) to exclude from compilation (partial, root-relative of full path)
    CompilerInputExcludedFiles : Optional[List[str]]
    # (optional) Pattern(s) to exclude from compilation
    CompilerInputExcludePattern : Optional[List[str]]
    # (optional) Explicit array of files to build
    CompilerInputFiles : Optional[List[str]]
    # (optional) Root path to use for .obj path generation for explicitly listed files
    CompilerInputFilesRoot : Optional[str]
    # (optional) Unity to build (or Unities)
    CompilerInputUnity : Optional[List[str]] 
    # (optional) ObjectList(s) whos output should be used as an input
    CompilerInputObjectLists : Optional[List[str]] 

    ###############################################
    # Cache & Distributed compilation control
    ###############################################
    # (optional) Allow caching of compiled objects if available (default true)
    AllowCaching : Optional[bool]
    # (optional) Allow distributed compilation if available (default true)
    AllowDistribution : Optional[bool]

    ###############################################
    # Custom preprocessor support
    ###############################################

    # (optional) Compiler to use for preprocessing
    Preprocessor : Optional[str]
    # (optional) Args to pass to compiler if using custom preprocessor
    PreprocessorOptions : Optional[str]

    ###############################################
    # Additional compiler options
    ###############################################
    # (optional) List of objects to be used with /FU
    CompilerForceUsing : Optional[List[str]]

    ###############################################
    # (optional) Properties to control precompiled header use
    ###############################################

    # (optional) Precompiled header (.cpp) file to compile
    PCHInputFile : Optional[str]
    # (optional) Precompiled header compilation output
    PCHOutputFile : Optional[str]
    # (optional) Options for compiler for precompiled header
    PCHOptions : Optional[str]

    ###############################################
    # Additional options
    ###############################################

    # (optional) Force targets to be built before this library (Rarely needed,
    # but useful when a library relies on generated code).
    PreBuildDependencies : Optional[List[str]]

    # (optional) Environment variables to use for local build
    # If set, librarian uses this environment
    # If not set, librarian uses .Environment from your Settings node
    Environment : Optional[List[str]]

    # (optional) Hide a target from -showtargets (default false)
    Hidden : Optional[bool]

    @staticmethod
    def keyword() -> str:
        return "Library"



from typing import List, Optional


class ProjectFileType:
    # FileType to use (e.g. "CppForm")
    FileType : str
    # Pattern of files to apply to (e.g. "*\\Forms\\*.h")
    Pattern  : str


class ProjectImport:
    # Condition for import
    Condition : str
    # Path to file to import
    Project   : str


class AdditonalConfigurationOptions:
    ###############################################
    # Compilation (optional)
    ###############################################

    # (optional) Command to launch when "build project" is selected.
    ProjectBuildCommand : Optional[str]
    # (optional) Command to launch when "rebuild project" is selected.
    ProjectRebuildCommand : Optional[str]
    # (optional) Command to launch when "clean project" is selected.
    ProjectCleanCommand : Optional[str]

    ###############################################
    # Compilation Input/Output (optional)
    ###############################################

    # (optional) Output generated by compilation.
    Output : Optional[str]
    # (optional) Output directory for Visual Studio.
    OutputDirectory : Optional[str]
    # (optional) Intermediate directory for Visual Studio.
    IntermediateDirectory : Optional[str]
    # (optional) Build log file for Visual Studio.
    BuildLogFile : Optional[str]
    # (optional) Directory to prepare Layout (for XboxOne).
    LayoutDir : Optional[str]
    # (optional) Files to filter from Layout (for XboxOne).
    LayoutExtensionFilter : Optional[str]

    ###############################################
    # Intellisense Options (optional)
    ###############################################

    # (optional) Preprocessor definitions.
    PreprocessorDefinitions : Optional[str]
    # (optional) Include search paths.
    IncludeSearchPath : Optional[str]
    # (optional) Force included files.
    ForcedIncludes : Optional[str]
    # (optional) Assembly search paths.
    AssemblySearchPath : Optional[str]
    # (optional) Forced Using assemblies.
    ForcedUsingAssemblies : Optional[str]
    # (optional) Additional compiler options.
    AdditionalOptions : Optional[str]

    ###############################################
    # Debugging Options (optional)
    ###############################################

    # (optional) Executable to launch when debugging.
    LocalDebuggerCommand : Optional[str]
    # (optional) Args passed to executable when debugging.
    LocalDebuggerCommandArguments : Optional[str]
    # (optional) Working Dir for executable when debugging.
    LocalDebuggerWorkingDirectory : Optional[str]
    # (optional) Environment variables when debugging.
    LocalDebuggerEnvironment : Optional[str]
    # (optional) Executable to launch when debugging remotely.
    RemoteDebuggerCommand : Optional[str]
    # (optional) Args passed to remote executable when debugging.
    RemoteDebuggerCommandArguments : Optional[str]
    # (optional) Working Dir for remote executable when debugging.
    RemoteDebuggerWorkingDirectory : Optional[str]
    # (optional) Debugger command for Xbox360 only.
    Xbox360DebuggerCommand : Optional[str]
    # (optional) Debugger flavor.
    DebuggerFlavor : Optional[str]
    # (optional) Aumid override (for XboxOne).
    AumidOverride : Optional[str]
    # (optional) Keyword
    Keyword : Optional[str]
    # (optional) Application type
    ApplicationType : Optional[str]
    # (optional) Application type revision
    ApplicationTypeRevision : Optional[str]
    # (optional) Set the Linux platform type
    TargetLinuxPlatform : Optional[str]
    # (optional) Set the project type GUID for Linux
    LinuxProjectType : Optional[str]
    # (optional) Path to package to debug
    PackagePath : Optional[str]
    # (optional) Additional symbol search paths for debugging
    AdditionalSymbolSearchPaths : Optional[str]

    ###############################################
    # Misc
    ###############################################

    # (optional) Specify PlatformToolset.
    PlatformToolset : Optional[str]
    # (optional) Specify deployment type for Xbox360.
    DeploymentType : Optional[str]
    # (optional) Specify files to add to deployment for Xbox360.
    DeploymentFiles : Optional[str]
    # (optional) Set RootNamespace for project
    RootNamespace : Optional[str]


class ProjectConfig(AdditonalConfigurationOptions):
    # Platform (e.g. Win32, X64, PS3 etc.)
    Platform : str
    # Config (e.g. Debug, Release etc.)
    Config   : str
    # (optional) Previously defined Build Target
    Target   : str


class VCXProject(AdditonalConfigurationOptions):
    # (optional) Alias
    Alias : Optional[str]

    ###############################################
    # Output Options
    ###############################################

    # Filename of project file
    ProjectOutput : str

    ###############################################
    # Input Options
    ###############################################

    # (optional) Paths to include in project
    ProjectInputPaths              : Optional[List[str]]
    # (optional) Paths to exclude from project
    ProjectInputPathsExclude       : Optional[List[str]]
    # (optional) Pattern(s) for files to exclude from project
    ProjectPatternToExclude        : Optional[List[str]]
    # (optional) Extensions to allow in path searches
    ProjectAllowedFileExtensions   : Optional[List[str]]
    # (optional) List of files to include in project
    ProjectFiles                   : Optional[List[str]]
    # (optional) List of files to exclude from project
    ProjectFilesToExclude          : Optional[List[str]]
    # (optional) Base path(s) for root folder(s) in project
    ProjectBasePath                : Optional[List[str]]
    # (optional) List of filters to override base file types (see above)
    ProjectFileTypes               : Optional[List[ProjectFileType]]

    ###############################################
    # Build Config Options
    ###############################################

    # (optional) List of project configurations, see below for details.
    # (default: Debug Win32, Release Win32, Debug X64, Release X64)
    ProjectConfigs                 : Optional[List[ProjectConfig]]

    ###############################################
    # Reference Options
    ###############################################

    # (optional) References to assemblies (e.g "System.Core")
    ProjectReferences              : Optional[List[str]]
    # (optional) References to projects (e.g. "myproj.csproj|{guid}")
    ProjectProjectReferences       : Optional[List[str]]

    ###############################################
    # Project Import Options
    ###############################################

    # (optional) Array of ProjectImports, see below for details
    ProjectProjectImports          : Optional[List[ProjectImport]]

    ###############################################
    # Other Options
    ###############################################

    # (optional) Override default generated ProjectGuid
    ProjectGuid                    : Optional[List[str]]
    # (optional) Default Language Property (for XboxOne/WinRT)
    DefaultLanguage                : Optional[List[str]]
    # (optional) Application Environment (for XboxOne/WinRT)
    ApplicationEnvironment         : Optional[List[str]]
    # (optional) Project will contain source control binding strings
    ProjectSccEntrySAK             : Optional[List[str]]

    @staticmethod
    def keyword() -> str:
        return "VCXProject"



from argparse import ArgumentParser
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext
from zetsubou.utils import logger
from zetsubou.utils.error_codes import EErrorCode, ReturnErrorcode


class CleanBuild(Command):
    @property
    def name(self):
        return 'clean_build'

    @property
    def desc(self):
        return 'Removes built binaries.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    @property
    def needs_project_loaded(self):
        return False

    def OnExecute(self, context: CommandContext):
        captured_exception = None
        dirs = ['build/bin', 'build/lib', 'build/obj']
        for dir in dirs:
            try:
                logger.Info(f"Removing built binaries from '{dir}' ...")
                if context.project_fs.exists(dir):
                    context.project_fs.removetree(dir)

            except Exception as err:
                logger.Error(f"Failed to remove '{dir}'")
                captured_exception = err

        if captured_exception is not None:
            raise ReturnErrorcode(EErrorCode.UNABLE_TO_CLEAN) from captured_exception



from enum import Enum, auto


class EArch(Enum):
    x64 = auto()
    x86 = auto()


class ESystem(Enum):
    Windows = auto()
    Linux = auto()


class EVersionSelector(Enum):
    # all = auto()
    latest = auto()
